""" route bundles of port (river routing)
"""

from typing import Callable, List, Optional
from numpy import float64, ndarray
import numpy as np
from pp.routing.connect import connect_strip
from pp.routing.connect import connect_elec_waypoints
from pp.routing.connect import connect_strip_way_points
from pp.routing.manhattan import generate_manhattan_waypoints

from pp.routing.u_groove_bundle import u_bundle_indirect
from pp.routing.u_groove_bundle import u_bundle_direct
from pp.routing.corner_bundle import corner_bundle
from pp.routing.path_length_matching import path_length_matched_points
from pp.name import autoname
from pp.component import ComponentReference, Component
from pp.port import Port
from pp.config import conf

METAL_MIN_SEPARATION = 10.0
BEND_RADIUS = conf.tech.bend_radius


def connect_bundle(
    start_ports,
    end_ports,
    route_filter=connect_strip_way_points,
    separation=5.0,
    bend_radius=BEND_RADIUS,
    extension_length=0,
    **kwargs,
):
    """ Connects bundle of ports using river routing.
    Chooses the correct u_bundle to use based on port angles

    Args:
        start_ports should all be facing in the same direction
        end_ports should all be facing in the same direction
        route_filter: function to connect
        separation: waveguide separation
        bend_radius: for the routes
        extension_length: adds waveguide extension

    """
    # Accept dict ot list
    if isinstance(start_ports, dict):
        start_ports = list(start_ports.values())

    if isinstance(end_ports, dict):
        end_ports = list(end_ports.values())

    nb_ports = len(start_ports)
    for p in start_ports:
        p.angle = int(p.angle) % 360

    for p in end_ports:
        p.angle = int(p.angle) % 360

    assert len(end_ports) == nb_ports
    assert (
        len(set([p.angle for p in start_ports])) <= 1
    ), "All start port angles should be the same"

    assert (
        len(set([p.angle for p in end_ports])) <= 1
    ), "All end port angles should be the same"

    # Ensure the correct bend radius is used
    def _route_filter(*args, **kwargs):
        kwargs["bend_radius"] = bend_radius
        return route_filter(*args, **kwargs)

    params = {
        "start_ports": start_ports,
        "end_ports": end_ports,
        "route_filter": _route_filter,
        "separation": separation,
        "bend_radius": bend_radius,
    }

    start_angle = start_ports[0].angle
    end_angle = end_ports[0].angle

    start_axis = "X" if start_angle in [0, 180] else "Y"
    end_axis = "X" if end_angle in [0, 180] else "Y"

    x_start = np.mean([p.x for p in start_ports])
    x_end = np.mean([p.x for p in end_ports])

    y_start = np.mean([p.y for p in start_ports])
    y_end = np.mean([p.y for p in end_ports])

    if start_axis == end_axis:
        if (
            start_angle == 0
            and end_angle == 180
            and x_start < x_end
            or start_angle == 180
            and end_angle == 0
            and x_start > x_end
            or start_angle == 90
            and end_angle == 270
            and y_start < y_end
            or start_angle == 270
            and end_angle == 90
            and y_start > y_end
        ):
            return link_ports(**params, **kwargs)

        elif start_angle == end_angle:
            return u_bundle_direct(**params, **kwargs)

        elif end_angle == (start_angle + 180) % 360:
            params["extension_length"] = extension_length
            return u_bundle_indirect(**params, **kwargs)
        else:
            raise NotImplementedError("This should never happen")

    else:
        return corner_bundle(**params, **kwargs)
        raise NotImplementedError("Routing along different axis not implemented yet")


def get_port_x(port: Port) -> float64:
    return port.midpoint[0]


def get_port_y(port):
    return port.midpoint[1]


def get_port_width(port):
    return port.width


def are_decoupled(x1, x1p, x2, x2p, sep=METAL_MIN_SEPARATION):
    if x2p + sep > x1:
        return False
    if x2 < x1p + sep:
        return False
    if x2 < x1p - sep:
        return False
    return True


def link_ports(
    start_ports: List[Port],
    end_ports: List[Port],
    separation: float = 5.0,
    route_filter: Callable = connect_strip_way_points,
    **routing_params,
) -> List[ComponentReference]:
    """Semi auto-routing for two lists of ports.

    Args:
        ports1: first list of ports
        ports2: second list of ports
        separation: minimum separation between two waveguides
        axis: specifies "X" or "Y"
              X (resp. Y) -> indicates that the ports should be sorted and
             compared using the X (resp. Y) axis
        bend_radius: If unspecified, attempts to get it from the waveguide definition of the first port in ports1
        route_filter: filter to apply to the manhattan waypoints
            e.g `connect_strip_way_points` for deep etch strip waveguide
        end_straight_offset: offset to add at the end of each waveguide
        sort_ports: * True -> sort the ports according to the axis.
                    * False -> no sort applied
        compute_array_separation_only: If True, returns the min distance which should be used between the two arrays instead of returning the connectors. Useful for budgeting space before instantiating other components.

    Returns:
        `[route_filter(r) for r in routes]` where routes is a list of lists of coordinates
        e.g with default `connect_strip_way_points`, returns a list of elements which can be added to a component


    The routing assumes manhattan routing between the different ports.
    The strategy is to modify `start_straight` and `end_straight` for each
    waveguide such that waveguides do not collide.

    .. code::

        1             X    X     X  X X  X
        |-----------|    |     |  | |  |-----------------------|
        |          |-----|     |  | |---------------|          |
        |          |          ||  |------|          |          |
        2 X          X          X          X          X          X


    start: at the top
    end: at the bottom

    The general strategy is:
    Group tracks which would collide together and apply the following method
    on each group:

        if x2 >= x1, increase ``end_straight``
            (as seen on the right 3 ports)
        otherwise, decrease ``end_straight``
            (as seen on the first 2 ports)

    We deal with negative end_straight by doing at the end
        end_straights = end_straights - min(end_straights)

    This method deals with different metal track/wg/wire widths too.

    """

    routes = link_ports_routes(
        start_ports,
        end_ports,
        separation=separation,
        routing_func=generate_manhattan_waypoints,
        **routing_params,
    )

    return [route_filter(route, **routing_params) for route in routes]


def link_ports_routes(
    start_ports: List[Port],
    end_ports: List[Port],
    separation: float,
    bend_radius: float = BEND_RADIUS,
    routing_func: Callable = generate_manhattan_waypoints,
    sort_ports: bool = True,
    end_straight_offset: Optional[float] = None,
    compute_array_separation_only: bool = False,
    verbose: int = 0,
    tol: float = 0.00001,
    **kwargs,
) -> List[ndarray]:
    """
    routing_func: Function used to connect two ports. Should be like `connect_strip`
    """
    ports1 = start_ports
    ports2 = end_ports
    if start_ports[0].angle in [0, 180]:
        axis = "X"
    else:
        axis = "Y"

    if len(ports1) == 0 or len(ports2) == 0:
        print("WARNING! Not linking anything, empty list of ports")
        return []

    if len(ports1) == 1 and len(ports2) == 1:
        if end_straight_offset:
            kwargs["end_straight"] = end_straight_offset
        return [
            routing_func(
                ports1[0],
                ports2[0],
                start_straight=0.05,
                bend_radius=bend_radius,
                **kwargs,
            )
        ]

    elems = []

    ## Contains end_straight of tracks which need to be adjusted together
    end_straights_in_group = []

    ## Once a group is finished, all the lengths are appended to end_straights
    end_straights = []

    ## Axis along which we sort the ports
    if axis in ["X", "x"]:
        f_key1 = get_port_y
        f_key2 = get_port_y
    else:
        f_key1 = get_port_x
        f_key2 = get_port_x

    if sort_ports:
        ports1.sort(key=f_key1)
        ports2.sort(key=f_key2)

    ## Keep track of how many ports should be routed together
    number_o_connectors_in_group = 0

    if axis in ["X", "x"]:
        x1_prev = get_port_y(ports1[0])
        x2_prev = get_port_y(ports2[0])
        y0 = get_port_x(ports2[0])
        y1 = get_port_x(ports1[0])
    else:  # X axis
        x1_prev = get_port_x(ports1[0])
        x2_prev = get_port_x(ports2[0])
        y0 = get_port_y(ports2[0])
        y1 = get_port_y(ports1[0])

    s = sign(y0 - y1)
    curr_end_straight = 0

    end_straight_offset = end_straight_offset or 15.0

    Le = end_straight_offset

    has_close_x_ports = False
    close_ports_thresh = 2 * bend_radius + 1.0

    ## First pass - loop on all the ports to find the tentative end_straights

    _w = get_port_width

    for i in range(len(ports1)):
        if axis in ["X", "x"]:
            x1 = get_port_y(ports1[i])
            x2 = get_port_y(ports2[i])
            y = get_port_x(ports2[i])
        else:
            x1 = get_port_x(ports1[i])
            x2 = get_port_x(ports2[i])
            y = get_port_y(ports2[i])

        dx = abs(x2 - x1)
        if dx < close_ports_thresh:
            has_close_x_ports = True

        """
        Compute the metal separation to use. This depends on the adjacent
        metal track widths
        """
        if i != len(ports1) - 1 and i != 0:
            # Deal with any track which is not on the edge
            max_width = max(_w(ports1[i + 1]), _w(ports1[i - 1]))
            curr_sep = 0.5 * (_w(ports1[i]) + max_width) + separation

        elif i == 0:
            # Deal with start edge case
            curr_sep = separation + 0.5 * (_w(ports1[0]) + _w(ports1[1]))

        elif i == len(ports1) - 1:
            # Deal with end edge case
            curr_sep = separation + 0.5 * (_w(ports1[-2]) + _w(ports1[-1]))

        if are_decoupled(x2, x2_prev, x1, x1_prev, sep=curr_sep):
            """
            If this metal track does not impact the previous one, then
            start a new group.
            """
            L = min(end_straights_in_group)
            end_straights += [max(x - L, 0) + Le for x in end_straights_in_group]

            # Start new group
            end_straights_in_group = []
            curr_end_straight = 0
            number_o_connectors_in_group = 0

        else:
            if x2 >= x1:
                curr_end_straight += curr_sep
            else:
                curr_end_straight -= curr_sep

        end_straights_in_group.append(curr_end_straight + (y - y0) * s)
        number_o_connectors_in_group += 1

        x1_prev = x1
        x2_prev = x2

    # Append the last group
    L = min(end_straights_in_group)

    end_straights += [max(x - L, 0) + Le for x in end_straights_in_group]

    if compute_array_separation_only:
        # If there is no port too close to each other in x, then there are
        # only two bends per route
        if not has_close_x_ports:
            return max(end_straights) + 2 * bend_radius
        else:
            return max(end_straights) + 4 * bend_radius

    ## Second pass - route the ports pairwise
    N = len(ports1)
    for i in range(N):
        if axis in ["X", "x"]:
            x1 = get_port_y(ports1[i])
            x2 = get_port_y(ports2[i])
        else:
            x1 = get_port_x(ports1[i])
            x2 = get_port_x(ports2[i])

        dx = abs(x2 - x1)

        # If both ports are aligned, we just need a straight line
        if dx < tol:
            elems += [
                routing_func(
                    ports1[i],
                    ports2[i],
                    start_straight=0,
                    end_straight=end_straights[i],
                    bend_radius=bend_radius,
                    **kwargs,
                )
            ]  #

        # Annoying case where it is too tight for direct manhattan routing
        elif dx < close_ports_thresh:

            a = close_ports_thresh + abs(dx)
            prt = ports1[i]
            angle = prt.angle
            dp_w = (0, -a) if axis in ["X", "x"] else (-a, 0)
            dp_e = (0, a) if axis in ["X", "x"] else (a, 0)
            do_step_aside = False
            if i == 0:
                ## If westest port, then we can safely step on the west further

                ## First check whether we have to step

                dx2 = ports1[i + 1].x - prt.x
                req_x = 2 * bend_radius + ports2[i].x - ports1[i].x

                if dx2 < req_x:
                    do_step_aside = True
                dp = dp_w

            elif i == N - 1:
                ## If eastest port, then we can safely step on the east further

                ## First check whether we have to step

                dx2 = prt.x - ports1[i - 1].x
                req_x = 2 * bend_radius + ports1[i].x - ports2[i].x

                if dx2 < req_x:
                    do_step_aside = True

                dp = dp_e

            else:
                ## Otherwise find closest port and escape where/if space permit

                dx1 = prt.x - ports1[i - 1].x
                dx2 = ports1[i + 1].x - prt.x
                do_step_aside = True
                if dx2 > dx1:
                    dp = dp_e
                else:
                    dp = dp_w

                ## If there is not enough space to step away, put a warning.
                ## This requires inspection on the mask. Raising an error
                ## would likely make it harder to debug. Here we will see
                ## a DRC error or an unwanted crossing on the mask.

                if max(dx1, dx2) < a:
                    print(
                        "WARNING - high risk of collision in routing. \
                    Ports too close to each other."
                    )
            _route = []
            if do_step_aside:
                tmp_port = prt.move_polar_copy(2 * bend_radius + 1.0, angle)
                tmp_port.move(dp)
                _route += [
                    routing_func(
                        prt, tmp_port.flip(), bend_radius=bend_radius, **kwargs
                    )
                ]

            else:
                tmp_port = prt
            if verbose > 2:
                print(
                    "STEPPING",
                    ports1[i].position,
                    tmp_port.position,
                    ports2[i].position,
                )
            _route += [
                routing_func(
                    tmp_port,
                    ports2[i],
                    start_straight=0.05,
                    end_straight=end_straights[i],
                    bend_radius=bend_radius,
                    **kwargs,
                )
            ]

            elems += _route
        # Usual case
        else:
            elems += [
                routing_func(
                    ports1[i],
                    ports2[i],
                    start_straight=0.05,
                    end_straight=end_straights[i],
                    bend_radius=bend_radius,
                    **kwargs,
                )
            ]
    return elems


def generate_waypoints_connect_bundle(*args, **kwargs):
    """
    returns a list of waypoints for each path generated with link_ports
    """
    return connect_bundle(*args, route_filter=lambda x, **params: x, **kwargs)


def compute_ports_max_displacement(start_ports, end_ports):
    if start_ports[0].angle in [0, 180]:
        a1 = [p.y for p in start_ports]
        a2 = [p.y for p in end_ports]
    else:
        a1 = [p.x for p in start_ports]
        a2 = [p.x for p in end_ports]

    return max(abs(max(a1) - min(a2)), abs(min(a1) - max(a2)))


def connect_bundle_path_length_match(
    ports1,
    ports2,
    separation=30.0,
    end_straight_offset=None,
    bend_radius=BEND_RADIUS,
    dL0=0,
    nb_loops=1,
    modify_segment_i=-2,
    route_filter=connect_strip_way_points,
    **kwargs,
):
    """
    Args:
        ports1,
        ports2,
        separation=30.0,
        end_straight_offset=None,
        bend_radius=BEND_RADIUS,
        dL0=0,
        nb_loops=1,
        modify_segment_i=-2,
        route_filter=connect_strip_way_points,
        **kwargs: extra arguments for inner call to generate_waypoints_connect_bundle

    Returns:
        [route_filter(l) for l in list_of_waypoints]

    """
    kwargs["separation"] = separation

    # Heuristic to get a correct default end_straight_offset to leave
    # enough space for path-length compensation

    if end_straight_offset is None:
        if modify_segment_i == -2:
            end_straight_offset = (
                compute_ports_max_displacement(ports1, ports2) / (2 * nb_loops)
                + separation
                + dL0
            )
        else:
            end_straight_offset = 0

    kwargs["end_straight_offset"] = end_straight_offset
    kwargs["bend_radius"] = bend_radius
    list_of_waypoints = generate_waypoints_connect_bundle(ports1, ports2, **kwargs)

    list_of_waypoints = path_length_matched_points(
        list_of_waypoints,
        dL0=dL0,
        bend_radius=bend_radius,
        nb_loops=nb_loops,
        modify_segment_i=modify_segment_i,
    )
    return [route_filter(waypoints) for waypoints in list_of_waypoints]


def link_electrical_ports(
    ports1: List[Port],
    ports2: List[Port],
    separation: float = METAL_MIN_SEPARATION,
    bend_radius: float = 0.0001,
    link_dummy_ports=False,
    route_filter: Callable = connect_elec_waypoints,
    **kwargs,
) -> List[ComponentReference]:
    """ Connect bundle of electrical ports

    Args:
        ports1: first list of ports
        ports2: second list of ports
        separation: minimum separation between two waveguides
        axis: specifies "X" or "Y"
              X (resp. Y) -> indicates that the ports should be sorted and
             compared using the X (resp. Y) axis
        bend_radius: If unspecified, attempts to get it from the waveguide definition of the first port in ports1
        route_filter: filter to apply to the manhattan waypoints
            e.g `connect_strip_way_points` for deep etch strip waveguide

        end_straight_offset: offset to add at the end of each waveguide
        sort_ports: * True -> sort the ports according to the axis.
                    * False -> no sort applied
        compute_array_separation_only: If True, returns the min distance which should be used between the two arrays instead of returning the connectors. Useful for budgeting space before instantiating other components.

    Returns:
        list of references of the electrical routes
    """
    if link_dummy_ports:
        new_ports1 = ports1
        new_ports2 = ports2

    else:

        def is_dummy(port):
            return hasattr(port, "is_dummy") and port.is_dummy

        to_keep1 = [not is_dummy(p) for p in ports1]
        to_keep2 = [not is_dummy(p) for p in ports2]
        to_keep = [x * y for x, y in zip(to_keep1, to_keep2)]

        new_ports1 = [p for i, p in enumerate(ports1) if to_keep[i]]
        new_ports2 = [p for i, p in enumerate(ports2) if to_keep[i]]

    return link_ports(
        new_ports1,
        new_ports2,
        separation,
        bend_radius=bend_radius,
        route_filter=route_filter,
        **kwargs,
    )


def link_optical_ports(
    ports1: List[Port],
    ports2: List[Port],
    separation: float = 5.0,
    route_filter: Callable = connect_strip_way_points,
    bend_radius: float = BEND_RADIUS,
    **kwargs,
) -> List[ComponentReference]:
    """ connect bundle of optical ports
    """
    return link_ports(
        ports1,
        ports2,
        separation,
        bend_radius=bend_radius,
        route_filter=route_filter,
        **kwargs,
    )


def sign(x):
    if x > 0:
        return 1
    else:
        return -1


def get_min_spacing(
    ports1: List[Port],
    ports2: List[Port],
    sep: float = 5.0,
    sort_ports: bool = True,
    radius: float = BEND_RADIUS,
) -> float:
    """
    Returns the minimum amount of spacing required to create a given fanout
    """

    if ports1[0].angle in [0, 180]:
        axis = "X"
    else:
        axis = "Y"

    j = 0
    min_j = 0
    max_j = 0
    if sort_ports:
        if axis in ["X", "x"]:
            ports1.sort(key=get_port_y)
            ports2.sort(key=get_port_y)
        else:
            ports1.sort(key=get_port_x)
            ports2.sort(key=get_port_x)

    for i in range(len(ports1)):
        if axis in ["X", "x"]:
            x1 = get_port_y(ports1[i])
            x2 = get_port_y(ports2[i])
        else:
            x1 = get_port_x(ports1[i])
            x2 = get_port_x(ports2[i])
        if x2 >= x1:
            j += 1
        else:
            j -= 1
        if j < min_j:
            min_j = j
        if j > max_j:
            max_j = j
    j = 0

    return (max_j - min_j) * sep + 2 * radius + 1.0


def link_optical_ports_no_grouping(
    ports1,
    ports2,
    sep=5.0,
    routing_func=connect_strip,
    radius=BEND_RADIUS,
    start_straight=None,
    end_straight=None,
    sort_ports=True,
):
    """
    Compared to link_ports, this function does not do any grouping.
    It is not as smart for the routing, but it can fall back on arclinarc
    connection if needed. We can also specify longer start_straight and end_straight

    Semi auto routing for optical ports
    The routing assumes manhattan routing between the different ports.
    The strategy is to modify ``start_straight`` and ``end_straight`` for each
    waveguide such that waveguides do not collide.


    We want to connect something like this:

    ::
         2             X    X     X  X X  X
           |-----------|    |     |  | |  |-----------------------|
           |          |-----|     |  | |---------------|          |
           |          |          ||  |------|          |          |
         1 X          X          X          X          X          X

    ``start`` is at the bottom
    ``end`` is at the top

    The general strategy is:

    if x2 < x1, decrease ``start straight``, and increase ``end_straight``
        (as seen on left two ports)
    otherwise, decrease ``start_straight``, and increase ``end_straight``
        (as seen on the last 3 right ports)

    Args:
        ports1: first list of optical ports
        ports2: second list of optical ports
        axis:   specifies "X" or "Y" direction along which the port is going
        routing_func:   ManhattanExpandedWgConnector or ManhattanWgConnector or any other connector function with the same input
        radius:         bend radius. If unspecified, uses the default radius
        start_straight: offset on the starting length before the first bend
        end_straight:   offset on the ending length after the last bend
        sort_ports:     True -> sort the ports according to the axis. False -> no sort applied

    Returns:
        a list of elements containing the connecting waveguides

    """

    if ports1[0].angle in [0, 180]:
        axis = "X"
    else:
        axis = "Y"

    elems = []
    j = 0

    # min and max offsets needed for avoiding collisions between waveguides
    min_j = 0
    max_j = 0

    if sort_ports:
        # Sort ports according to X or Y
        if axis in ["X", "x"]:
            ports1.sort(key=get_port_y)
            ports2.sort(key=get_port_y)
        else:
            ports1.sort(key=get_port_x)
            ports2.sort(key=get_port_x)

    # Compute max_j and min_j
    for i in range(len(ports1)):
        if axis in ["X", "x"]:
            x1 = ports1[i].position.y
            x2 = ports2[i].position.y
        else:
            x1 = ports1[i].position.x
            x2 = ports2[i].position.x
        if x2 >= x1:
            j += 1
        else:
            j -= 1
        if j < min_j:
            min_j = j
        if j > max_j:
            max_j = j
    j = 0

    if start_straight is None:
        start_straight = 0.2

    if end_straight is None:
        end_straight = 0.2

    start_straight += max_j * sep
    end_straight += -min_j * sep

    # Do case with wire direct if the ys are close to each other
    for i in range(len(ports1)):

        if axis in ["X", "x"]:
            x1 = ports1[i].position.y
            x2 = ports2[i].position.y
        else:
            x1 = ports1[i].position.x
            x2 = ports2[i].position.x

        s_straight = start_straight - j * sep
        e_straight = j * sep + end_straight

        if radius is None:
            elems += [
                routing_func(
                    ports1[i],
                    ports2[i],
                    start_straight=s_straight,
                    end_straight=e_straight,
                )
            ]
        else:
            elems += [
                routing_func(
                    ports1[i],
                    ports2[i],
                    start_straight=s_straight,
                    end_straight=e_straight,
                    bend_radius=radius,
                )
            ]

        if x2 >= x1:
            j += 1
        else:
            j -= 1
    return elems


@autoname
def test_connect_bundle():

    xs_top = [-100, -90, -80, 0, 10, 20, 40, 50, 80, 90, 100, 105, 110, 115]

    pitch = 127.0
    N = len(xs_top)
    xs_bottom = [(i - N / 2) * pitch for i in range(N)]

    top_ports = [Port("top_{}".format(i), (xs_top[i], 0), 0.5, 270) for i in range(N)]

    bottom_ports = [
        Port("bottom_{}".format(i), (xs_bottom[i], -400), 0.5, 90) for i in range(N)
    ]

    top_cell = Component(name="connect_bundle")
    elements = connect_bundle(top_ports, bottom_ports)
    for e in elements:
        top_cell.add(e)
    top_cell.name = "connect_bundle"

    return top_cell


@autoname
def test_connect_corner(N=6, config="A"):

    d = 10.0

    sep = 5.0
    top_cell = Component(name="connect_corner")

    if config in ["A", "B"]:
        a = 100.0
        ports_A_TR = [
            Port("A_TR_{}".format(i), (d, a / 2 + i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_TL = [
            Port("A_TL_{}".format(i), (-d, a / 2 + i * sep), 0.5, 180) for i in range(N)
        ]
        ports_A_BR = [
            Port("A_BR_{}".format(i), (d, -a / 2 - i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_BL = [
            Port("A_BL_{}".format(i), (-d, -a / 2 - i * sep), 0.5, 180)
            for i in range(N)
        ]

        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port("B_TR_{}".format(i), (a / 2 + i * sep, d), 0.5, 90) for i in range(N)
        ]
        ports_B_TL = [
            Port("B_TL_{}".format(i), (-a / 2 - i * sep, d), 0.5, 90) for i in range(N)
        ]
        ports_B_BR = [
            Port("B_BR_{}".format(i), (a / 2 + i * sep, -d), 0.5, 270) for i in range(N)
        ]
        ports_B_BL = [
            Port("B_BL_{}".format(i), (-a / 2 - i * sep, -d), 0.5, 270)
            for i in range(N)
        ]

        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    elif config in ["C", "D"]:
        a = N * sep + 2 * d
        ports_A_TR = [
            Port("A_TR_{}".format(i), (a, d + i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_TL = [
            Port("A_TL_{}".format(i), (-a, d + i * sep), 0.5, 180) for i in range(N)
        ]
        ports_A_BR = [
            Port("A_BR_{}".format(i), (a, -d - i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_BL = [
            Port("A_BL_{}".format(i), (-a, -d - i * sep), 0.5, 180) for i in range(N)
        ]

        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port("B_TR_{}".format(i), (d + i * sep, a), 0.5, 90) for i in range(N)
        ]
        ports_B_TL = [
            Port("B_TL_{}".format(i), (-d - i * sep, a), 0.5, 90) for i in range(N)
        ]
        ports_B_BR = [
            Port("B_BR_{}".format(i), (d + i * sep, -a), 0.5, 270) for i in range(N)
        ]
        ports_B_BL = [
            Port("B_BL_{}".format(i), (-d - i * sep, -a), 0.5, 270) for i in range(N)
        ]

        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    if config in ["A", "C"]:
        for ports1, ports2 in zip(ports_A, ports_B):
            elements = connect_bundle(ports1, ports2)
            top_cell.add(elements)

    elif config in ["B", "D"]:
        for ports1, ports2 in zip(ports_A, ports_B):
            elements = connect_bundle(ports2, ports1)
            top_cell.add(elements)

    return top_cell


@autoname
def test_connect_bundle_udirect(dy=200, angle=270):

    xs1 = [-100, -90, -80, -55, -35, 24, 0] + [200, 210, 240]

    axis = "X" if angle in [0, 180] else "Y"

    pitch = 10.0
    N = len(xs1)
    xs2 = [50 + i * pitch for i in range(N)]

    if axis == "X":
        ports1 = [Port("top_{}".format(i), (0, xs1[i]), 0.5, angle) for i in range(N)]

        ports2 = [
            Port("bottom_{}".format(i), (dy, xs2[i]), 0.5, angle) for i in range(N)
        ]

    else:
        ports1 = [Port("top_{}".format(i), (xs1[i], 0), 0.5, angle) for i in range(N)]

        ports2 = [
            Port("bottom_{}".format(i), (xs2[i], dy), 0.5, angle) for i in range(N)
        ]

    top_cell = Component(name="connect_bundle_udirect")
    elements = connect_bundle(ports1, ports2)
    for e in elements:
        top_cell.add(e)

    return top_cell


@autoname
def test_connect_bundle_u_indirect(dy=-200, angle=180):

    xs1 = [-100, -90, -80, -55, -35] + [200, 210, 240]

    axis = "X" if angle in [0, 180] else "Y"

    pitch = 10.0
    N = len(xs1)
    xs2 = [50 + i * pitch for i in range(N)]

    a1 = angle
    a2 = a1 + 180

    if axis == "X":
        ports1 = [Port("top_{}".format(i), (0, xs1[i]), 0.5, a1) for i in range(N)]

        ports2 = [Port("bottom_{}".format(i), (dy, xs2[i]), 0.5, a2) for i in range(N)]

    else:
        ports1 = [Port("top_{}".format(i), (xs1[i], 0), 0.5, a1) for i in range(N)]

        ports2 = [Port("bottom_{}".format(i), (xs2[i], dy), 0.5, a2) for i in range(N)]

    top_cell = Component("connect_bundle_u_indirect")
    elements = connect_bundle(ports1, ports2)
    for e in elements:
        top_cell.add(e)

    return top_cell


@autoname
def test_facing_ports():
    dy = 200.0
    xs1 = [-500, -300, -100, -90, -80, -55, -35, 200, 210, 240, 500, 650]

    pitch = 10.0
    N = len(xs1)
    xs2 = [-20 + i * pitch for i in range(N // 2)]
    xs2 += [400 + i * pitch for i in range(N // 2)]

    a1 = 90
    a2 = a1 + 180

    ports1 = [Port("top_{}".format(i), (xs1[i], 0), 0.5, a1) for i in range(N)]
    ports2 = [Port("bottom_{}".format(i), (xs2[i], dy), 0.5, a2) for i in range(N)]

    top_cell = Component("test_facing_ports")
    elements = connect_bundle(ports1, ports2)
    # elements = link_ports_path_length_match(ports1, ports2)
    top_cell.add(elements)

    return top_cell


def demo_connect_bundle():
    """ combines all the connect_bundle tests """

    y = 400.0
    x = 500
    y0 = 900
    dy = 200.0
    c = Component("connect_bundle")
    for j, s in enumerate([-1, 1]):
        for i, angle in enumerate([0, 90, 180, 270]):
            _cmp = test_connect_bundle_u_indirect(dy=s * dy, angle=angle)
            _cmp_ref = _cmp.ref(position=(i * x, j * y))
            c.add(_cmp_ref)

            _cmp = test_connect_bundle_udirect(dy=s * dy, angle=angle)
            _cmp_ref = _cmp.ref(position=(i * x, j * y + y0))
            c.add(_cmp_ref)

    for i, config in enumerate(["A", "B", "C", "D"]):
        _cmp = test_connect_corner(config=config)
        _cmp_ref = _cmp.ref(position=(i * x, 1700))
        c.add(_cmp_ref)

    _cmp = test_facing_ports()
    _cmp_ref = _cmp.ref(position=(800, 1820))
    c.add(_cmp_ref)

    return c


def demo_connect_bundle_small(bend_radius=5):
    import pp

    c = pp.c.mmi1x2()
    elements = connect_bundle([c.ports["E0"]], [c.ports["E1"]], bend_radius=5)
    c.add(elements)
    return c


if __name__ == "__main__":
    import pp

    c = demo_connect_bundle()
    # c = demo_connect_bundle_small()
    pp.show(c)



"""
This is a sample on how to define custom components.
You can make a repo out of this file, having one custom component per file
"""
import os
import shutil
import pytest
import pp
from pp.config import CONFIG
from pp.autoplacer.yaml_placer import place_from_yaml
from pp.components.spiral_inner_io import spiral_inner_io_euler
from pp.add_termination import add_gratings_and_loop_back
from pp.routing.connect import connect_strip_way_points
from pp.add_padding import add_padding_to_grid
from pp.generate_does import generate_does

from pp.mask.merge_metadata import merge_metadata


def _route_filter(*args, **kwargs):
    return connect_strip_way_points(
        *args, taper_factory=None, start_straight=5.0, end_straight=5.0, **kwargs
    )


def add_te(component, **kwargs):
    c = pp.routing.add_fiber_array(
        component,
        grating_coupler=pp.c.grating_coupler_elliptical_te,
        route_filter=_route_filter,
        **kwargs,
    )
    c.test = "passive_optical_te"
    c = add_padding_to_grid(c)
    return c


def add_tm(component, **kwargs):
    c = pp.routing.add_fiber_array(
        component,
        grating_coupler=pp.c.grating_coupler_elliptical_tm,
        route_filter=_route_filter,
        bend_radius=20,
        **kwargs,
    )
    c = add_padding_to_grid(c)
    return c


@pp.autoname
def coupler_te(gap, length, wg_width=0.5, nominal_wg_width=0.5):
    """ sample of component cutback """
    c = pp.c.coupler(wg_width=wg_width, gap=gap, length=length)
    cc = add_te(c)
    return cc


@pp.autoname
def spiral_te(wg_width=0.5, length=2):
    """ sample of component cutback

    Args:
        wg_width: um
        lenght: mm
    """
    c = spiral_inner_io_euler(wg_width=wg_width, length=length)
    cc = add_gratings_and_loop_back(
        component=c,
        grating_coupler=pp.c.grating_coupler_elliptical_te,
        bend_factory=pp.c.bend_circular,
    )
    return cc


@pp.autoname
def spiral_tm(wg_width=0.5, length=2):
    """ sample of component cutback """
    c = spiral_inner_io_euler(wg_width=wg_width, length=length, dx=10, dy=10, N=5)
    cc = add_gratings_and_loop_back(
        component=c,
        grating_coupler=pp.c.grating_coupler_elliptical_tm,
        bend_factory=pp.c.bend_circular,
    )
    return cc


component_type2factory = dict(
    spiral_te=spiral_te, spiral_tm=spiral_tm, coupler_te=coupler_te
)


@pytest.fixture
def cleandir():
    build_folder = CONFIG["samples_path"] / "mask_custom" / "build"
    if build_folder.exists():
        shutil.rmtree(build_folder)


@pytest.fixture
def chdir():
    workspace_folder = CONFIG["samples_path"] / "mask_custom"
    os.chdir(workspace_folder)


@pytest.mark.usefixtures("cleandir")
def test_mask(precision=2e-9):
    workspace_folder = CONFIG["samples_path"] / "mask_custom"
    build_path = workspace_folder / "build"
    doe_root_path = build_path / "cache_doe"
    doe_metadata_path = build_path / "doe"
    mask_path = build_path / "mask"
    does_yml = workspace_folder / "does.yml"

    mask_path.mkdir(parents=True, exist_ok=True)

    gdspath = mask_path / "sample_mask.gds"
    markdown_path = gdspath.with_suffix(".md")
    json_path = gdspath.with_suffix(".json")
    test_metadata_path = gdspath.with_suffix(".tp.json")

    generate_does(
        str(does_yml),
        component_type2factory=component_type2factory,
        precision=precision,
        doe_root_path=doe_root_path,
        doe_metadata_path=doe_metadata_path,
    )

    top_level = place_from_yaml(does_yml, precision=precision, root_does=doe_root_path)
    top_level.write(str(gdspath))

    merge_metadata(gdspath=gdspath)

    assert gdspath.exists()
    assert markdown_path.exists()
    assert json_path.exists()
    assert test_metadata_path.exists()

    report = open(markdown_path).read()
    assert report.count("#") == 2, f" only {report.count('#')} DOEs in {markdown_path}"

    return gdspath


if __name__ == "__main__":
    # from pprint import pprint
    # pprint(component_type2factory)
    c = test_mask()
    pp.klive.show(c)
    # c = coupler_te(gap=0.3, length=20)
    # pp.show(c)



from typing import Dict, List
from pp.port import Port


def flip(port: Port) -> Port:
    """ Flip a phidl Port """
    return Port(port.name, port.midpoint, port.width, port.orientation + 180)


def direction_ports_from_list_ports(optical_ports: List[Port]) -> Dict[str, List[Port]]:
    direction_ports = {x: [] for x in ["E", "N", "W", "S"]}
    for p in optical_ports:
        p.angle = (p.angle + 360.0) % 360
        if p.angle <= 45.0 or p.angle >= 315:
            direction_ports["E"].append(p)
        elif p.angle <= 135.0 and p.angle >= 45.0:
            direction_ports["N"].append(p)
        elif p.angle <= 225.0 and p.angle >= 135.0:
            direction_ports["W"].append(p)
        else:
            direction_ports["S"].append(p)

    for direction, list_ports in list(direction_ports.items()):
        if direction in ["E", "W"]:
            list_ports.sort(key=lambda p: p.y)

        if direction in ["S", "N"]:
            list_ports.sort(key=lambda p: p.x)

    return direction_ports


def check_ports_have_equal_spacing(list_ports):
    if not list_ports:
        raise ValueError("list_ports should not be empty")

    angle = get_list_ports_angle(list_ports)
    if angle in [0, 180]:
        xys = [p.y for p in list_ports]
    else:
        xys = [p.x for p in list_ports]

    seps = [round(abs(c2 - c1), 5) for c1, c2 in zip(xys[1:], xys[:-1])]
    different_seps = set(seps)
    if len(different_seps) > 1:
        raise ValueError(
            "Ports should have the same separation. \
        Got {}".format(
                different_seps
            )
        )

    return different_seps.pop()


def get_list_ports_angle(list_ports):
    if not list_ports:
        return None
    if len(set([p.angle for p in list_ports])) > 1:
        raise ValueError(
            "All port angles should be the same.\
        Got {}".format(
                list_ports
            )
        )
    return list_ports[0].orientation



import pp
from pp.container import container
from pp.components.taper import taper


def add_taper_elements(component, taper=taper):
    """returns ports and taper elements for a component"""
    taper = pp.call_if_func(taper)
    ports = []
    elements = []
    c = pp.Component()

    for port_name, port in component.ports.copy().items():
        if port.port_type == "optical":
            taper_ref = c << pp.call_if_func(taper)
            taper_ref.connect(taper_ref.ports["2"].name, port)
            elements.append(taper_ref)
            ports.append(taper_ref.ports["1"])
    return ports, elements


@container
@pp.port.deco_rename_ports
def add_tapers(component, taper=taper, suffix="t", port_type="optical"):
    """returns component optical tapers for component """

    taper = pp.call_if_func(taper)
    c = pp.Component(name=f"{component.name}_{suffix}")

    for port_name, port in component.ports.copy().items():
        if port.port_type == port_type:
            taper_ref = c << pp.call_if_func(taper)
            taper_ref.connect(taper_ref.ports["2"].name, port)
            c.add_port(name=port_name, port=taper_ref.ports["1"])
        else:
            c.add_port(name=port_name, port=port)
    c.add_ref(component)
    return c


if __name__ == "__main__":
    c = pp.c.waveguide(width=2)
    t = pp.c.taper(width2=2)

    # cc = add_tapers(component=c, taper=t, suffix="t")
    # print(cc.ports.keys())
    # print(cc.settings.keys())
    # pp.show(cc)

    ports, elements = add_taper_elements(component=c, taper=t)
    c.ports = ports
    c.add(elements)
    pp.show(c)
    print(c.ports)



from typing import Callable
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_tm

import pp
from pp.add_tapers import add_tapers
from pp.components.taper import taper
from pp.container import container

from pp.routing.route_fiber_array import route_fiber_array
from pp.routing.get_input_labels import get_input_labels
from pp.component import Component


def add_fiber_array_te(*args, **kwargs):
    return add_fiber_array(*args, **kwargs)


def add_fiber_array_tm(*args, grating_coupler=grating_coupler_tm, **kwargs):
    return add_fiber_array(*args, grating_coupler=grating_coupler, **kwargs)


@container
def add_fiber_array(
    component: Component,
    grating_coupler: Component = grating_coupler_te,
    gc_port_name: str = "W0",
    component_name: None = None,
    taper_factory: Callable = taper,
    taper_length: float = 10.0,
    get_route_factory: Callable = route_fiber_array,
    get_input_labels_function: Callable = get_input_labels,
    **kwargs,
) -> Component:
    """returns component with optical IO (tapers, south routes and grating_couplers)

    Args:
        component: to connect
        grating_coupler: grating coupler instance, function or list of functions
        bend_factory: bend_circular
        straight_factory: waveguide
        fanout_length: None  # if None, automatic calculation of fanout length
        max_y0_optical: None
        with_align_ports: True, adds loopback structures
        waveguide_separation: 4.0
        bend_radius: BEND_RADIUS
        list_port_labels: None, adds TM labels to port indices in this list
        connected_port_list_ids: None # only for type 0 optical routing
        nb_optical_ports_lines: 1
        force_manhattan: False
        excluded_ports:
        grating_indices: None
        routing_waveguide: None
        routing_method: connect_strip
        gc_port_name: W0
        optical_routing_type: None: autoselection, 0: no extension, 1: standard, 2: check
        gc_rotation: -90
        layer_label: LAYER.LABEL
        input_port_indexes: [0]
        component_name: for the label
        taper_factory: taper function
        get_route_factory: route_fiber_array

    .. plot::
      :include-source:

       import pp
       from pp.routing import add_fiber_array

       c = pp.c.crossing()
       cc = add_fiber_array(c)
       pp.plotgds(cc)

    """
    c = component
    if not c.ports:
        return c

    if isinstance(grating_coupler, list):
        gc = grating_coupler[0]
    else:
        gc = grating_coupler
    gc = pp.call_if_func(gc)

    gc_polarization = gc.polarization

    component_name = component_name or c.name
    name = f"{component_name}_{gc_polarization}"
    cc = pp.Component(name=name)

    port_width_gc = gc.ports[gc_port_name].width

    optical_ports = c.get_ports_list(port_type="optical")
    port_width_component = optical_ports[0].width

    if port_width_component != port_width_gc:
        c = add_tapers(
            c,
            taper_factory(
                length=taper_length, width1=port_width_gc, width2=port_width_component
            ),
        )

    # for pn, p in c.ports.items():
    #     print(p.name, p.port_type, p.layer)

    elements, io_gratings_lines, _ = get_route_factory(
        component=c,
        grating_coupler=grating_coupler,
        gc_port_name=gc_port_name,
        component_name=component_name,
        get_input_labels_function=get_input_labels_function,
        **kwargs,
    )
    if len(elements) == 0:
        return c

    for e in elements:
        cc.add(e)
    for io_gratings in io_gratings_lines:
        cc.add(io_gratings)
    cc.add(c.ref())
    cc.move(origin=io_gratings_lines[0][0].ports[gc_port_name], destination=(0, 0))

    for pname, p in c.ports.items():
        if p.port_type != "optical":
            cc.add_port(pname, port=p)

    return cc


def test_type0():
    component = pp.c.coupler(gap=0.244, length=5.67)
    cc = add_fiber_array(component, optical_routing_type=0)
    pp.write_gds(cc)
    return cc


def test_type1():
    component = pp.c.coupler(gap=0.2, length=5.0)
    cc = add_fiber_array(component, optical_routing_type=1)
    pp.write_gds(cc)
    return cc


def test_type2():
    c = pp.c.coupler(gap=0.244, length=5.67)
    c.polarization = "tm"
    cc = add_fiber_array(c, optical_routing_type=2)
    pp.write_gds(cc)
    return cc


def demo_tapers():
    c = pp.c.waveguide(width=2)
    cc = add_fiber_array(c, optical_routing_type=2)
    return cc


def demo_te_and_tm():
    c = pp.Component()
    w = pp.c.waveguide()
    wte = add_fiber_array(w, grating_coupler=pp.c.grating_coupler_elliptical_te)
    wtm = add_fiber_array(w, grating_coupler=pp.c.grating_coupler_elliptical_tm)
    c.add_ref(wte)
    wtm_ref = c.add_ref(wtm)
    wtm_ref.movey(wte.size_info.height)
    return c


if __name__ == "__main__":
    gcte = pp.c.grating_coupler_te
    gctm = pp.c.grating_coupler_tm

    # from pprint import pprint

    layer_label = pp.LAYER.TEXT
    layer_label = (66, 5)

    # cc = demo_tapers()
    # cc = test_type1()
    # pprint(cc.get_json())

    # c = pp.c.coupler(gap=0.2, length=5.6)

    c = pp.c.mmi2x2()
    # c = pp.c.waveguide()

    c.y = 0
    cc = add_fiber_array(
        c,
        # optical_routing_type=0,
        # optical_routing_type=1,
        # optical_routing_type=2,
        # layer_label=layer_label,
        # get_route_factory=route_fiber_single,
        # get_route_factory=route_fiber_array,
        grating_coupler=[gcte, gctm, gcte, gctm],
    )
    # cc = demo_te_and_tm()
    # print(cc.ports.keys())
    pp.show(cc)
    print(cc.get_settings()["component"])



""" CD SEM structures
"""

import pp

from pp.components.bend_circular import bend_circular
from pp.components.bend_circular import bend_circular_trenches

from pp.components.waveguide import waveguide_trenches
from pp.layers import LAYER
from pp.port import rename_ports_by_orientation


def square_middle(side=0.5, layer=LAYER.WG):
    component = pp.Component()
    a = side / 2
    component.add_polygon([(-a, -a), (a, -a), (a, a), (-a, a)], layer=layer)
    return component


def text(t="U"):
    return pp.c.text(text=t, layer=pp.LAYER.WG, size=5)


CENTER_SHAPES_MAP = {"S": square_middle, "U": text("U"), "D": text("L")}


@pp.autoname
def cdsem_straight(w, dw, spacing=5.0, length=20.0):
    """
    w
    dw
    """

    c = pp.Component()

    c.move(c.size_info.cc, (0, 0))
    return c


@pp.autoname
def cdsem_straight_density(
    wg_width=0.372, trench_width=0.304, x=500, y=50.0, margin=2.0
):
    """ horizontal grating etch lines

    TE: 676nm pitch, 304nm gap, 372nm line
    TM: 1110nm pitch, 506nm gap, 604nm line

    Args:
        w: wg_width
        s: trench_width
    """
    c = pp.Component()
    period = wg_width + trench_width
    n_o_lines = int((y - 2 * margin) / period)
    length = x - 2 * margin

    slab = pp.c.rectangle_centered(x=x, y=y, layer=LAYER.WG)
    slab_ref = c.add_ref(slab)
    c.absorb(slab_ref)

    tooth = pp.c.rectangle_centered(x=length, y=trench_width, layer=LAYER.SLAB150)

    for i in range(n_o_lines):
        _tooth = c.add_ref(tooth)
        _tooth.movey((-n_o_lines / 2 + 0.5 + i) * period)
        c.absorb(_tooth)

    c.move(c.size_info.cc, (0, 0))
    return c


@pp.autoname
def cdsem_target(width_center=0.5):
    radii = [5.0, 10.0]
    c = pp.Component()
    a = 1.0
    w = 3 * a / 4
    c.add_ref(square_middle())
    ctr = c.add_ref(square_middle())
    ctl = c.add_ref(square_middle())
    cbr = c.add_ref(square_middle())
    cbl = c.add_ref(square_middle())

    ctr.move((w, w))
    ctl.move((-w, w))
    cbr.move((w, -w))
    cbl.move((-w, -w))

    w_min = width_center * 0.92
    w0 = width_center
    w_max = width_center * 1.08

    for radius in radii:
        b = a + radius
        _b_tr = bend_circular(radius=radius, width=w0)
        b_tr = _b_tr.ref(position=(b, a), rotation=90, port_id="W0")

        _b_bl = bend_circular(radius=radius, width=w0)
        b_bl = _b_bl.ref(position=(-b, -a), rotation=270, port_id="W0")

        _b_br = bend_circular(radius=radius, width=w_max)
        b_br = _b_br.ref(position=(a, -b), rotation=0, port_id="W0")

        _b_tl = bend_circular(radius=radius, width=w_min)
        b_tl = _b_tl.ref(position=(-a, b), rotation=180, port_id="W0")

        c.add([b_tr, b_tl, b_bl, b_br])

    return c


@pp.autoname
def cdsem_uturn(
    width=0.5, cladding_offset=3.0, radius=10, symbol_bot="S", symbol_top="D"
):
    c = pp.Component()
    r = radius

    # bend90 = bend_circular_deep_rib(width=width, radius=r, cladding_offset=cladding_offset)
    # wg = wg_deep_rib(width=width, length=2 * r)

    bend90 = bend_circular_trenches(width=width, radius=r)
    wg = waveguide_trenches(width=width, length=2 * r)

    # bend90.ports()
    rename_ports_by_orientation(bend90)

    # Add the U-turn on waveguide layer
    b1 = c.add_ref(bend90)
    b2 = c.add_ref(bend90)

    b2.connect("N0", b1.ports["W0"])

    wg1 = c.add_ref(wg)
    wg1.connect("W0", b1.ports["N0"])

    wg2 = c.add_ref(wg)
    wg2.connect("W0", b2.ports["W0"])

    # Add symbols

    sym1 = c.add_ref(CENTER_SHAPES_MAP[symbol_bot]())
    sym1.movey(r)
    sym2 = c.add_ref(CENTER_SHAPES_MAP[symbol_top])
    sym2.movey(2 * r)

    c.move(c.size_info.cc, (0, 0))
    return c


if __name__ == "__main__":
    # c = cdsem_straight_density()
    c = cdsem_uturn()
    pp.show(c)

    # c = pcm_bend()
    # pp.write_gds(c)
    # pp.show(c)



from typing import Any, List, Tuple, Union
from phidl.device_layout import Label
import pp
from pp.routing.route_fiber_array import route_fiber_array
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.component import Component, ComponentReference


def route_fiber_single(
    component: Component,
    optical_io_spacing: int = 50,
    grating_coupler: Component = grating_coupler_te,
    min_input2output_spacing: int = 230,
    optical_routing_type: int = 1,
    optical_port_labels: None = None,
    excluded_ports: List[Any] = [],
    **kwargs
) -> Union[
    Tuple[List[Union[ComponentReference, Label]], List[List[ComponentReference]], None],
    Tuple[
        List[Union[ComponentReference, Label]],
        List[Union[List[ComponentReference], ComponentReference]],
        None,
    ],
]:
    """Returns component I/O for optical testing with single input and oputput fibers (no fiber array)

    Args:
        component: to add grating couplers
        optical_io_spacing: between grating couplers
        grating_coupler:
        min_input2output_spacing: so opposite fibers do not touch
        optical_routing_type: 0, 1, 2
        optical_port_labels: port labels that need connection
        excluded_ports: ports excluded from routing

    """
    component = component.copy()
    component_copy = component.copy()

    if optical_port_labels is None:
        optical_ports = component.get_ports_list(port_type="optical")
    else:
        optical_ports = [component.ports[lbl] for lbl in optical_port_labels]
    optical_ports = [p for p in optical_ports if p.name not in excluded_ports]
    N = len(optical_ports)

    if isinstance(grating_coupler, list):
        grating_couplers = [pp.call_if_func(g) for g in grating_coupler]
        grating_coupler = grating_couplers[0]
    else:
        grating_coupler = pp.call_if_func(grating_coupler)
        grating_couplers = [grating_coupler] * N

    gc_port2center = getattr(grating_coupler, "port2center", grating_coupler.xsize / 2)
    if component.xsize + 2 * gc_port2center < min_input2output_spacing:
        fanout_length = (
            pp.drc.snap_to_grid(
                min_input2output_spacing - component.xsize - 2 * gc_port2center, 10
            )
            / 2
        )
    else:
        fanout_length = None

    """
         _________
        |         |_E1
     W0_|         |
        |         |_E0
        |_________|

    rotate +90 deg and route West ports to South

          E1  E0
         _|___|_
        |       |
        |       |
        |       |
        |       |
        |       |
        |       |
        |_______|
            |
            W0

    """
    # route west ports to south
    component = component.rotate(90)
    west_ports = component.get_ports_dict(prefix="W")
    north_ports = {
        p.name: p for p in component.ports.values() if not p.name.startswith("W")
    }
    component.ports = west_ports

    elements_south, gratings_south, _ = route_fiber_array(
        component=component,
        with_align_ports=False,
        optical_io_spacing=optical_io_spacing,
        fanout_length=fanout_length,
        grating_coupler=grating_couplers[0],
        optical_routing_type=optical_routing_type,
        **kwargs
    )

    # route north ports
    component = component_copy.rotate(-90)
    north_ports = {
        p.name: p for p in component.ports.values() if not p.name.startswith("W")
    }
    component.ports = north_ports

    elements_north, gratings_north, _ = route_fiber_array(
        component=component,
        with_align_ports=False,
        optical_io_spacing=optical_io_spacing,
        fanout_length=fanout_length,
        grating_coupler=grating_couplers[1:],
        **kwargs
    )
    for e in elements_north:
        elements_south.append(e.rotate(180))

    if len(gratings_north) > 0:
        for io in gratings_north[0]:
            gratings_south.append(io.rotate(180))

    return elements_south, gratings_south, None


if __name__ == "__main__":
    gcte = pp.c.grating_coupler_te
    gctm = pp.c.grating_coupler_tm

    # c = pp.c.crossing()
    # c = pp.c.mmi2x2()
    # c = pp.c.ring_double()  # FIXME
    c = pp.c.cross(length=500)
    c = pp.c.waveguide(width=2, length=500)

    elements, gc, _ = route_fiber_single(c, grating_coupler=[gcte, gctm, gcte, gctm])

    cc = pp.Component()
    cr = cc << c.rotate(90)
    cr.x = 0
    cr.y = 0

    for e in elements:
        cc.add(e)
    for e in gc:
        cc.add(e)
    pp.show(cc)



from typing import List, Union
import numpy as np

import pp
from pp.container import container
from omegaconf.listconfig import ListConfig
from phidl.device_layout import Layer
from pp.component import Component


@container
def add_padding(
    component: Component,
    padding: Union[float, int] = 50,
    x: None = None,
    y: None = None,
    layers: Union[List[ListConfig], List[Layer]] = [pp.LAYER.PADDING],
    suffix: str = "p",
) -> Component:
    """ adds padding layers to a NEW component that has the same:
    - ports
    - settings
    - test_protocols and data_analysis_protocols

    as the old component
    """
    x = x if x is not None else padding
    y = y if y is not None else padding

    c = pp.Component(name=f"{component.name}_{suffix}")
    c << component
    c.ports = component.ports

    points = [
        [c.xmin - x, c.ymin - y],
        [c.xmax + x, c.ymin - y],
        [c.xmax + x, c.ymax + y],
        [c.xmin - x, c.ymax + y],
    ]
    for layer in layers:
        c.add_polygon(points, layer=layer)
    return c


@container
def add_padding_to_grid(
    component,
    grid_size=127,
    x=10,
    y=10,
    bottom_padding=5,
    layers=[pp.LAYER.PADDING],
    suffix="p",
):
    """ returns component width a padding layer on each side
    matches a minimum size
    """
    c = pp.Component(name=f"{component.name}_{suffix}")
    c << component
    c.ports = component.ports

    if c.size_info.height < grid_size:
        y_padding = grid_size - c.size_info.height
    else:
        n_grids = np.ceil(c.size_info.height / grid_size)
        y_padding = n_grids * grid_size - c.size_info.height

    if c.size_info.width < grid_size:
        x_padding = grid_size - c.size_info.width
    else:
        n_grids = np.ceil(c.size_info.width / grid_size)
        x_padding = n_grids * grid_size - c.size_info.width

    x_padding -= x
    y_padding -= y

    points = [
        [c.xmin - x_padding / 2, c.ymin - bottom_padding],
        [c.xmax + x_padding / 2, c.ymin - bottom_padding],
        [c.xmax + x_padding / 2, c.ymax + y_padding - bottom_padding],
        [c.xmin - x_padding / 2, c.ymax + y_padding - bottom_padding],
    ]
    for layer in layers:
        c.add_polygon(points, layer=layer)
    return c


if __name__ == "__main__":
    c = pp.c.waveguide(length=128)
    cc = add_padding(component=c, layers=[(2, 0)], suffix="p")
    # cc = add_padding_to_grid(c, layers=[(2, 0)])
    # cc = add_padding_to_grid(c)
    print(cc.settings)
    print(cc.ports)
    pp.show(cc)



""" You can remap layers

"""


import pp

c = pp.Component("waveguides_sample")


wg1 = c << pp.c.waveguide(length=10, width=1)
wg2 = c << pp.c.waveguide(length=10, width=2, layer=pp.LAYER.SLAB90)
wg3 = c << pp.c.waveguide(length=10, width=3, layer=pp.LAYER.SLAB150)

wg2.connect(port="W0", destination=wg1.ports["E0"])
wg3.connect(port="W0", destination=wg2.ports["E0"], overlap=1)


c.remap_layers({pp.LAYER.WG: pp.LAYER.SLAB150})

pp.qp(c)  # quickplot it!
pp.show(c)  # show it in klayout



""" write a Component from the YAML netlist

Deprecated! use pp/component_from_yaml instead!

.. code::

             top_arm
        -CP1=       =CP2-
             bot_arm
"""

from typing import Union, IO, Any
import pathlib
import io
from omegaconf import OmegaConf

from pp.component import Component
from pp.components import component_type2factory as component_type2factory_default
from pp.netlist_to_gds import netlist_to_component


sample = """

instances:
    CP1:
      component: mmi1x2
      settings:
          width_mmi: 4.5
          length_mmi: 10
    CP2:
        component: mmi1x2
        settings:
            width_mmi: 4.5
            length_mmi: 5
        transformations: mirror_y
    arm_top:
        component: mzi_arm
        settings:
            L0: 10
            DL: 0
    arm_bot:
        component: mzi_arm
        settings:
            L0: 100
            DL: 0
        transformations: mirror_x

connections:
    - [CP1, E0, arm_bot, W0]
    - [arm_bot, E0, CP2, E0]
    - [CP1, E1, arm_top, W0]
    - [arm_top, E0, CP2, E0]

ports_map:
    W0: [CP1, W0]
    E0: [CP2, W0]
    E_TOP_0: [arm_top, E_0]
    E_TOP_1: [arm_top, E_1]
    E_TOP_2: [arm_top, E_2]
    E_TOP_3: [arm_top, E_3]
    E_BOT_0: [arm_bot, E_0]
    E_BOT_1: [arm_bot, E_1]
    E_BOT_2: [arm_bot, E_2]
    E_BOT_3: [arm_bot, E_3]

"""


def netlist_from_yaml(
    yaml: Union[str, pathlib.Path, IO[Any]], component_type2factory=None,
) -> Component:
    """ Loads Component settings from YAML file, and connections

    Deprecated! use component_from_yaml instead

    Args:
        netlist: YAML IO describing instances, connections and ports_map

    Returns:
        Component

    .. code-block:: yaml

        instances:
            CP1:
              component: mmi1x2
              settings:
                  width_mmi: 4.5
                  length_mmi: 10
            CP2:
                component: mmi1x2
                settings:
                    width_mmi: 4.5
                    length_mmi: 5
                transformations: mirror_y
            arm_top:
                component: mzi_arm
                settings:
                    L0: 10
                    DL: 0
            arm_bot:
                component: mzi_arm
                settings:
                    L0: 100
                    DL: 0
                transformations: mirror_x

        ports_map:
            W0: [CP1, W0]
            E0: [CP2, W0]
            E_TOP_0: [arm_top, E_0]
            E_TOP_1: [arm_top, E_1]
            E_TOP_2: [arm_top, E_2]
            E_TOP_3: [arm_top, E_3]
            E_BOT_0: [arm_bot, E_0]
            E_BOT_1: [arm_bot, E_1]
            E_BOT_2: [arm_bot, E_2]
            E_BOT_3: [arm_bot, E_3]

        connections:
            - [CP1, E0, arm_bot, W0]
            - [arm_bot, E0, CP2, E0]
            - [CP1, E1, arm_top, W0]
            - [arm_top, E0, CP2, E0]

    """

    yaml = io.StringIO(yaml) if isinstance(yaml, str) and "\n" in yaml else yaml
    conf = OmegaConf.load(yaml)
    component_type2factory = component_type2factory or component_type2factory_default

    instances = {}
    for instance_name in conf.instances:
        instance_conf = conf.instances[instance_name]
        component_type = instance_conf["component"]
        component_settings = instance_conf["settings"] or {}
        instance = component_type2factory[component_type](**component_settings)
        instance_transformations = instance_conf["transformations"] or "None"
        instance_properties = instance_conf["properties"] or {}
        for k, v in instance_properties.items():
            setattr(instance, k, v)
        instance.name = instance_name
        instances[instance_name] = (instance, instance_transformations)

    connections = conf.connections
    ports_map = conf.ports_map
    return netlist_to_component(instances, connections, ports_map)


def test_netlist_from_yaml():
    c = netlist_from_yaml(sample)
    assert len(c.get_dependencies()) == 4
    return c


if __name__ == "__main__":
    import pp

    # c = test_netlist_from_yaml()
    c = netlist_from_yaml(sample)
    pp.show(c)



import itertools
import pp
from pp.components.coupler import coupler
from pp.components.crossing_waveguide import crossing45
from pp.components.crossing_waveguide import compensation_path
from pp.routing.repackage import package_optical2x2
from pp.port import get_ports_facing
from pp.config import GRID_PER_UNIT

COUNTER = itertools.count()


def gen_tmp_port_name():
    return "{}".format(next(COUNTER))


def swap(waveguides, i, j):
    a = waveguides[i]
    waveguides[i] = waveguides[j]
    waveguides[j] = a
    return waveguides


def dist(i, wgs1, wgs2):
    a = wgs1.index(i)
    b = wgs2.index(i)
    return b - a


def get_sequence_cross(
    waveguides_start, waveguides_end, iter_max=100, symbols=["X", "-"]
):
    """
    Args:
        waveguides_start : list of the input port indices
        waveguides_end : list of the output port indices
        symbols : [`X` , `S`]
        symbols to be used in the returned sequence:
        `X`: represents the crossing symbol: two Xs next
            to each-other means that the two modes have to be swapped
        `S`: Straight waveguide, or compensation path typically

    Returns:
        sequence of crossings to achieve the permutations between two columns of I/O
    """
    wgs = list(waveguides_start)
    waveguides_end = list(waveguides_end)
    N = len(wgs)
    sequence = []
    X, S = symbols  # Cross, Straight symbols
    nb_iters = 0
    while wgs != waveguides_end:
        if nb_iters > iter_max:
            print(
                "Exceeded max number of iterations. The following I/O are mismatched:"
            )
            for i in range(len(waveguides_end)):
                print(wgs[i], "<->", waveguides_end[i])
            return sequence

        if nb_iters > 2 and sequence[-1] == sequence[-2]:
            print("Two consecutive sequences are the same. Got stuck")
            return sequence

        swaps = []
        i = 0
        total_dist = 0
        while i < N - 1:
            a = wgs[i]
            b = wgs[i + 1]
            d1 = dist(a, wgs, waveguides_end)
            d2 = dist(b, wgs, waveguides_end)
            total_dist += abs(d1) + abs(d2)

            """
            # The equality cases are very important:
            # if one waveguide needs to cross, then even if the other one is
            # already at the right place, it must swap to allow the other one
            # to cross
            """

            if d1 >= 0 and d2 <= 0 and not (d1 == 0 and d2 == 0):
                wgs = swap(wgs, i, i + 1)
                swaps += [X, X]
                i += 1
                # We cannot swap twice the same waveguide on the same iteration, so we skip the next waveguide by incrementing

                # Edge case: Cannot swap if only one wg left so it has to be a straight
                if i == N - 2:
                    swaps += [S]
            else:
                # Edge case if only one wg remain it is straight
                swaps += [S]
                if i == N - 2:
                    swaps += [S]
            i += 1

        sequence.append(swaps)
        nb_iters += 1
    return sequence


def component_sequence_to_str(sequence):
    """
    Transform a sequence of components (such as the one obtained from
    get_sequence_cross_str) into an ASCII block which can be used either as
    a cartoon or as an input for component_lattice(lattice = ...)
    """
    component_txt_lattice = ""
    M = len(sequence[0])

    for i in range(M):
        j = M - 1 - i
        line = ""
        for col in sequence:
            line += col[j]
        line += "\n"
        component_txt_lattice += line

    return component_txt_lattice


def get_sequence_cross_str(waveguides_start, waveguides_end, iter_max=100):
    seq = get_sequence_cross(
        waveguides_start, waveguides_end, iter_max=iter_max, symbols=["X", "-"]
    )

    return component_sequence_to_str(seq)


def component_lattice(
    lattice="""
        C-X
        CXX
        CXX
        C-X
        """,
    components=None,
    name="lattice",
):
    """
    A lattice of N inputs and outputs with components at given locations
    Columns must have components with the same x spacing between ports input/output ports
    Lines must have components with the same y spacing between input and output ports

    Lattice example:

    .. code::

        X-X
        XCX
        XCX
        X-X

    .. plot::
      :include-source:

      import pp
      from pp.routing.repackage import package_optical2x2
      from pp.components.crossing_waveguide import crossing45
      from pp.components.crossing_waveguide import compensation_path

      components =  {
            "C": package_optical2x2(component=pp.c.coupler(), port_spacing=40.0),
            "X": crossing45(port_spacing=40.0),
            "-": compensation_path(crossing45=crossing45(port_spacing=40.0)),
      }
      c = pp.c.component_lattice(components=components)
      pp.plotgds(c)

    """
    components = components or {
        "C": package_optical2x2(component=coupler(), port_spacing=40.0),
        "X": crossing45(port_spacing=40.0),
        "-": compensation_path(crossing45=crossing45(port_spacing=40.0)),
    }

    # Find y spacing and check that all components have same y spacing

    y_spacing = None
    for cmp in components.values():
        # cmp = pp.call_if_func(cmp)

        for direction in ["W", "E"]:
            ports_dir = get_ports_facing(cmp.ports, direction)
            ports_dir.sort(key=lambda p: p.y)
            nb_ports = len(ports_dir)
            if nb_ports > 1:
                _y_spacing = (ports_dir[-1].y - ports_dir[0].y) / (nb_ports - 1)
                if y_spacing is None:
                    y_spacing = _y_spacing
                else:
                    assert abs(y_spacing - _y_spacing) < 0.1 / GRID_PER_UNIT, (
                        "All component must have the same y port spacing. Got"
                        f" {y_spacing}, {_y_spacing} for {cmp.name}"
                    )

    a = y_spacing
    columns, columns_to_length = parse_lattice(lattice, components)
    keys = sorted(columns.keys())

    components_to_nb_input_ports = {}
    for c in components.keys():
        components_to_nb_input_ports[c] = len(get_ports_facing(components[c], "W"))

    component = pp.Component(name)
    x = 0
    for i in keys:
        col = columns[i]
        j = 0
        L = columns_to_length[i]
        skip = 0  # number of lines to skip depending on the number of ports
        for c in col:
            y = -j * a
            if skip == 1:
                j += skip
                skip = 0
                continue

            if c in components.keys():
                # Compute the number of ports to skip: They will already be
                # connected since they belong to this component

                nb_inputs = components_to_nb_input_ports[c]
                skip = nb_inputs - 1
                _cmp = components[c].ref((x, y), port_id="W{}".format(skip))
                component.add(_cmp)

                if i == 0:
                    _ports = get_ports_facing(_cmp, "W")
                    for _p in _ports:
                        component.add_port(gen_tmp_port_name(), port=_p)

                if i == keys[-1]:
                    _ports = get_ports_facing(_cmp, "E")
                    for _p in _ports:
                        component.add_port(gen_tmp_port_name(), port=_p)

            else:
                raise ValueError(
                    "component symbol {} is not part of                 components"
                    " dictionnary".format(c)
                )

            j += 1
        x += L

    component = pp.port.rename_ports_by_orientation(component)
    return component


def parse_lattice(lattice, components):
    # extract each column
    lines = lattice.replace(" ", "").split("\n")
    columns = {}
    columns_to_length = {}
    for line in lines:
        if len(line) > 0:
            i = 0
            for c in line:
                if i not in columns.keys():
                    columns[i] = []

                columns[i].append(c)
                if c in components.keys():
                    cmp = components[c]
                    columns_to_length[i] = cmp.ports["E0"].x - cmp.ports["W0"].x

                i += 1

    return columns, columns_to_length


if __name__ == "__main__":
    components = {
        "C": package_optical2x2(component=pp.c.coupler, port_spacing=40.0),
        "X": crossing45(port_spacing=40.0),
        "-": compensation_path(crossing45=crossing45(port_spacing=40.0)),
    }
    c = pp.c.component_lattice(components=components)
    pp.show(c)



"""

When we decorate a function with autoname we can also pass a flat `with_ports` that will add port markers to our component

"""
import pp
from pp.add_pins import add_pins_triangle


def test_pins():
    c = pp.c.waveguide(pins=True)
    return c


def test_pins_custom():
    """ We can even define the `pins_function` that we use to add markers to each port

    """
    c = pp.c.waveguide(pins=True, pins_function=add_pins_triangle)
    return c


if __name__ == "__main__":
    c = test_pins()
    c = test_pins_custom()
    pp.qp(c)
    pp.show(c)



import numpy as np

import pp
from pp.components.pcm.cd import CENTER_SHAPES_MAP
from pp.components.pcm.cd import square_middle

from pp.components.bend_circular import bend_circular
from pp.components.bend_circular import bend_circular_ridge
from pp.components.bend_circular import bend_circular_slot

from pp.components.waveguide import waveguide
from pp.components.waveguide import waveguide_ridge
from pp.components.waveguide import waveguide_slot


@pp.autoname
def cd_bend(
    L=2.0,
    radius=2.0,
    width=0.4,
    center_shapes="SU",
    bend90_factory=pp.c.bend_circular,
    waveguide_factory=pp.c.waveguide,
):
    """ bends and straights connected together
    for CDSEM measurement
    """

    component = pp.Component()
    _straight = waveguide_factory(length=L, width=width)
    _bend = bend90_factory(radius=radius, width=width)

    straight1 = _straight.ref(rotation=90, port_id="W0")
    component.add(straight1)

    bend1 = component.add_ref(_bend)
    bend1.connect(port="N0", destination=straight1.ports["E0"])

    straight2 = component.add_ref(_straight)
    straight2.connect(port="W0", destination=bend1.ports["W0"])

    bend2 = component.add_ref(_bend)
    bend2.connect(port="N0", destination=straight2.ports["E0"])

    # Add center shapes.
    # Center the first shape in the list
    # Then stack the others underneath

    # If we do ridge waveguides, add a slab
    markers_with_slabs = bend90_factory == bend_circular_ridge

    center = np.array([radius + L / 2, L / 2])
    center_shape_side = 0.4 if markers_with_slabs else 0.5
    center_shape_spacing = 0.2
    sep = center_shape_side + center_shape_spacing
    for cs_name in center_shapes:
        _shape_func = CENTER_SHAPES_MAP[cs_name]
        _shape = _shape_func(side=center_shape_side)
        _shape_ref = _shape.ref(position=center)
        component.add(_shape_ref)
        component.absorb(_shape_ref)

        # If with slabs, add the square slab on top of each marker
        if markers_with_slabs:
            _shape = square_middle(side=0.5)
            _shape_ref = _shape.ref(position=center)
            component.add(_shape_ref)
            component.absorb(_shape_ref)

        center += (0, -sep)

    return component


@pp.autoname
def cd_bend_strip(**kwargs):
    return cd_bend(**kwargs, bend90_factory=bend_circular, waveguide_factory=waveguide)


@pp.autoname
def cd_bend_rib(cladding=0.3, **kwargs):
    def bend(*a, **k):
        return bend_circular_ridge(*a, cladding=cladding, **k)

    def wg(*a, **k):
        return waveguide_ridge(*a, cladding=cladding, **k)

    return cd_bend(**kwargs, bend90_factory=bend, waveguide_factory=wg)


@pp.autoname
def cd_bend_slot(gap=0.2, **kwargs):
    def bend(*a, **k):
        return bend_circular_slot(*a, gap=gap, **k)

    def wg(*a, **k):
        return waveguide_slot(*a, gap=gap, **k)

    return cd_bend(**kwargs, bend90_factory=bend, waveguide_factory=wg)


if __name__ == "__main__":
    c = cd_bend_strip(width=0.46)
    pp.show(c)



"""
Sometimes, when a component is mostly composed of sub-components adjacent to each other, it can be easier to define the component by:

- instances
- placements
- routes
- ports: exposed ports in the new components

"""

import pp


@pp.autoname
def test_netlist_yaml():
    """

    .. code::

           arm_top
            _____
           |     |
       CP1=       =CP2=
           |_____|

            arm_bot
   """

    yaml = """
    instances:
        mmi_long:
          component: mmi1x2
          settings:
            width_mmi: 4.5
            length_mmi: 10
        mmi_short:
          component: mmi1x2
          settings:
            width_mmi: 4.5
            length_mmi: 5

    placements:
        mmi_long:
            rotation: 180
            x: 100
            y: 100

    routes:
        mmi_short,E1: mmi_long,E0

    ports:
        E0: mmi_short,W0
        W0: mmi_long,W0
    """

    c = pp.component_from_yaml(yaml)
    return c


if __name__ == "__main__":
    c = test_netlist_yaml()
    pp.show(c)
    pp.plotgds(c)



import pp
from pp.components import bend_euler180
from pp.components.component_sequence import component_sequence


@pp.autoname
def cutback_component(
    component,
    cols=4,
    rows=5,
    bend_radius=10,
    port1_id="W0",
    port2_id="E0",
    middle_couples=2,
):
    """ Flips the component, good for tapers that end in wide waveguides
    Args:
        component
        cols
        rows

    """
    bend180 = bend_euler180(radius=bend_radius)

    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "A": (component, port1_id, port2_id),
        "B": (component, port2_id, port1_id),
        "D": (bend180, "W0", "W1"),
        "C": (bend180, "W1", "W0"),
    }

    # Generate the sequence of staircases

    s = ""
    for i in range(rows):
        s += "AB" * cols
        s += "D" if i % 2 == 0 else "C"

    s = s[:-1]
    s += "AB" * middle_couples

    for i in range(rows):
        s += "AB" * cols
        s += "D" if (i + rows) % 2 == 0 else "C"

    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(s, string_to_device_in_out_ports)
    c.update_settings(n_devices=len(s))
    return c


@pp.autoname
def cutback_component_flipped(
    component,
    cols=4,
    rows=5,
    bend_radius=10,
    port1_id="E0",
    port2_id="W0",
    middle_couples=2,
):
    bend180 = bend_euler180(radius=bend_radius)

    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "A": (component, port1_id, port2_id),
        "B": (component, port2_id, port1_id),
        "D": (bend180, "W0", "W1"),
        "C": (bend180, "W1", "W0"),
    }

    # Generate the sequence of staircases

    s = ""
    for i in range(rows):
        s += "AB" * cols
        s += "C" if i % 2 == 0 else "D"

    s = s[:-1]
    s += "AB" * middle_couples

    for i in range(rows):
        s += "AB" * cols
        s += "D" if (i + rows + 1) % 2 == 0 else "C"

    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(s, string_to_device_in_out_ports)
    c.update_settings(n_devices=len(s))
    return c


@pp.autoname
def cutback_polarization_rotator(n_devices_target, design=3):
    """ sample of component cutback """
    rows = 4
    cols = n_devices_target // (rows * 2)
    from pp.components.polarization_rotator import polarization_rotator

    c = cutback_component(
        component=polarization_rotator(design=design), rows=rows, cols=cols
    )
    cc = pp.routing.add_fiber_array(
        c,
        grating_coupler=pp.c.grating_coupler_elliptical_tm,
        optical_routing_type=0,
        connected_port_list_ids=["out", "in"],
    )

    return cc


if __name__ == "__main__":
    pass



""" functions to manage different GDS libraries
"""

import os
import shutil
from pp.write_component import write_component
import pp
from pp.config import CONFIG, conf
from gdsdiff.gdsdiff import gdsdiff

GDS_PATH_TMP = CONFIG["gdslib_test"]


def get_library_path(tech=conf.tech.name):
    return CONFIG["gdslib"] / tech


def get_component_path(component, tech=conf.tech.name):
    return get_library_path(tech) / (component.name + ".gds")


def is_same_as_in_library(
    component, tech=conf.tech.name, gdspath_tmp=CONFIG["gdslib_test"],
):
    """
    component: a pp.Component
    tech: the tech where to find a reference component
    """

    # # Check if the component is here
    # if not component_is_in_library(component, tech):
    # return False

    # Library component path
    lib_component_dir = get_library_path(tech)

    lib_hash = None
    lib_component = pp.load_component(component.name, lib_component_dir)

    lib_hash = lib_component.hash_geometry()
    new_component_hash = component.hash_geometry()
    return new_component_hash == lib_hash


def component_is_in_library(component, tech=conf.tech.name):
    lib_component_path = get_component_path(component, tech)
    return os.path.exists(lib_component_path)


def _copy_component(src, dest):
    for ext in [".gds", ".ports", ".json"]:
        shutil.copy(src.with_suffix(ext), dest.with_suffix(ext))


def add_component(
    component,
    tech,
    add_ports_to_all_cells=False,
    with_confirmation=True,
    force_replace=False,
    add_port_pins=True,
):
    """
    Arg:
        component <pp.Component>
        tech <str>

    Add a component to a given tech
    """

    lib_component_path = get_component_path(component, tech)
    # If it is a new library, create the path
    dirname = os.path.dirname(lib_component_path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    # Save component to make sure we have all the port and json files
    new_component_path = GDS_PATH_TMP / (component.name + ".gds")

    params = {
        "add_port_pins": add_port_pins,
        "add_ports_to_all_cells": add_ports_to_all_cells,
    }

    write_component(component, new_component_path, **params)

    # Reload component (to make sure ports info is added etc...)
    component = pp.load_component(component.name, GDS_PATH_TMP)
    write_component(component, new_component_path, **params)

    # If component is not in library, simply add it
    if not component_is_in_library(component, tech) or force_replace:
        _copy_component(new_component_path, lib_component_path)
        # write_component(component, lib_component_path,
        # add_ports_to_all_cells=add_ports_to_all_cells)
        return

    # If it is already there, compare the hash
    if is_same_as_in_library(component, tech):

        # We can safely overwrite the component since the geometry has not changed
        if not hasattr(component, "properties"):
            component.properties = {}
        _copy_component(new_component_path, lib_component_path)
        print(
            "{} cell geometry unchanged (cell/port names may have changed).".format(
                component.name
            )
        )

    elif with_confirmation:
        # Show the diff
        try:
            gds_diff = gdsdiff(lib_component_path, component)
            pp.show(gds_diff)

            a = input(
                "Do you want to replace {} with new version? [Y/N]".format(
                    component.name
                )
            )
            if a == "Y":
                _copy_component(new_component_path, lib_component_path)
                print("Replaced component {} by new one".format(component.name))
            else:
                print("Did not modify the component")
        except Exception as e:
            print("Cannot show the diff")
            print(e)
    else:
        _copy_component(new_component_path, lib_component_path)
        print("Replaced component {} by new one".format(component.name))



import pathlib
import numpy as np
from pp.component import Component
import pp


data_path = pathlib.Path(__file__).parent / "csv_data"


def grating_coupler_uniform_optimized(
    widths,
    width_grating=11,
    length_taper=150,
    width=0.5,
    partial_etch=False,
    layer=1,
    layer_partial_etch=2,
    taper=None,
):
    """ Grating coupler uniform (not focusing)

    Args:
        widths: of each teeth
        width_grating: 11
        length_taper: 150
        width: 0.5
        partial_etch: False

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_uniform()
      pp.plotgds(c)

    """
    # returns a fiber grating
    c = Component()
    x = 0

    if partial_etch:
        partetch_overhang = 5

        # make the etched areas (opposite to teeth)
        for i, wt in enumerate(widths):
            if i % 2 == 1:
                _compass = pp.c.compass(
                    size=[wt, width_grating + partetch_overhang * 2],
                    layer=layer_partial_etch,
                )
                cgrating = c.add_ref(_compass)
                cgrating.x += x + wt / 2
            x += wt

        # draw the deep etched square around the grating
        xgrating = np.sum(widths)
        deepbox = c.add_ref(pp.c.compass(size=[xgrating, width_grating], layer=layer))
        deepbox.movex(xgrating / 2)
    else:
        for i, wt in enumerate(widths):
            if i % 2 == 0:
                cgrating = c.add_ref(
                    pp.c.compass(size=[wt, width_grating], layer=layer)
                )
                cgrating.x += x + wt / 2
            x += wt

    # make the taper
    if taper is None:
        taper = pp.c.taper(
            length=length_taper,
            width1=width,
            width2=width_grating,
            port=None,
            layer=layer,
        )
    taper_ref = c.add_ref(taper)
    taper_ref.xmax = 0
    port = taper_ref.ports.get("W0") or taper_ref.ports.get("1")
    c.add_port(port=taper_ref.ports[port.name], name="E0")
    return c


@pp.autoname
def grating_coupler_uniform_1etch_h220_e70(**kwargs):
    csv_path = data_path / "grating_coupler_1etch_h220_e70.csv"
    import pandas as pd

    d = pd.read_csv(csv_path)
    return grating_coupler_uniform_optimized(
        widths=d["widths"], partial_etch=True, **kwargs
    )


@pp.autoname
def grating_coupler_uniform_2etch_h220_e70(**kwargs):
    csv_path = data_path / "grating_coupler_2etch_h220_e70_e220.csv"
    import pandas as pd

    d = pd.read_csv(csv_path)
    return grating_coupler_uniform_optimized(
        widths=d["widths"], partial_etch=True, **kwargs
    )


@pp.autoname
def grating_coupler_uniform_1etch_h220_e70_taper_w11_l200(**kwargs):
    from pp.components.taper_from_csv import taper_w11_l200

    taper = taper_w11_l200()
    return grating_coupler_uniform_1etch_h220_e70(taper=taper)


@pp.autoname
def grating_coupler_uniform_1etch_h220_e70_taper_w10_l200(**kwargs):
    from pp.components.taper_from_csv import taper_w10_l200

    taper = taper_w10_l200()
    return grating_coupler_uniform_1etch_h220_e70(taper=taper, width_grating=10)


@pp.autoname
def grating_coupler_uniform_1etch_h220_e70_taper_w10_l100(**kwargs):
    from pp.components.taper_from_csv import taper_w10_l100

    taper = taper_w10_l100()
    return grating_coupler_uniform_1etch_h220_e70(taper=taper, width_grating=10)


if __name__ == "__main__":
    # widths = [0.3, 0.5, 0.3]
    # c = grating_coupler_uniform_optimized(widths=widths, partial_etch=False)
    # c = grating_coupler_uniform_optimized(widths=widths, partial_etch=True)

    # c = grating_coupler_uniform_1etch_h220_e70()
    c = grating_coupler_uniform_2etch_h220_e70()
    # c = grating_coupler_uniform_1etch_h220_e70_taper_w11_l200()
    # c = grating_coupler_uniform_1etch_h220_e70_taper_w10_l200()
    # c = grating_coupler_uniform_1etch_h220_e70_taper_w10_l100()
    pp.show(c)



""" Write component GDS + metadata
write_component_type: try to load a component from library or creates if it does not exist
write_component: write component and metadata
"""

import pathlib
import json
from pathlib import PosixPath
from typing import Optional
from phidl import device_layout as pd

from pp.config import CONFIG, conf
from pp.name import get_component_name
from pp.components import component_type2factory
from pp import klive
from pp.component import Component

from pp.layers import LAYER


def get_component_type(
    component_type, component_type2factory=component_type2factory, **kwargs
):
    """ returns a component from the factory """
    component_name = get_component_name(component_type, **kwargs)
    return component_type2factory[component_type](name=component_name, **kwargs)


def write_component_type(
    component_type,
    overwrite=False,
    path_directory=CONFIG["gds_directory"],
    component_type2factory=component_type2factory,
    **kwargs,
):
    """ write_component by type or function

    Args:
        component_type: can be function or factory name
        overwrite:
        path_directory: to store GDS + metadata
        component_type2factory: factory dictionary
        **kwargs: component args
    """
    if callable(component_type):
        component_type = component_type.__name__

    assert type(component_type) == str

    component_name = get_component_name(component_type, **kwargs)
    gdspath = path_directory / (component_name + ".gds")
    path_directory.mkdir(parents=True, exist_ok=True)

    if not gdspath.exists() or overwrite:
        component = component_type2factory[component_type](
            name=component_name, **kwargs
        )
        component.type = component_type
        write_component(component, gdspath)

    return gdspath


def write_component_report(component, json_path=None):
    """ write component GDS and metadata:

    Args:
        component:
        json_path
    """

    json_path = json_path or CONFIG["gds_directory"] / component.name + ".json"

    ports_path = json_path[:-5] + ".ports"

    """ write ports """
    if len(component.ports) > 0:
        with open(ports_path, "w") as fw:
            for _, port in component.ports.items():
                fw.write(
                    "{}, {:.3f}, {:.3f}, {}, {:.3f}, {}\n".format(
                        port.name,
                        port.x,
                        port.y,
                        int(port.orientation),
                        port.width,
                        port.layer,
                    )
                )

    """ write json """
    with open(json_path, "w+") as fw:
        fw.write(json.dumps(component.get_json(), indent=2))
    return json_path


def write_component(
    component: Component,
    gdspath: Optional[PosixPath] = None,
    path_library: PosixPath = CONFIG["gds_directory"],
    precision: float = 1e-9,
    settings: None = None,
    with_settings_label: bool = conf.tech.with_settings_label,
) -> str:
    """ write component GDS and metadata:

    - gds
    - ports
    - properties

    Args:
        component:
        gdspath:
        path_library
        precision: to save GDS points
        settings: dict of settings
    """

    gdspath = gdspath or path_library / (component.name + ".gds")
    gdspath = pathlib.Path(gdspath)
    ports_path = gdspath.with_suffix(".ports")
    json_path = gdspath.with_suffix(".json")

    """ write GDS """
    gdspath = write_gds(
        component=component,
        gdspath=str(gdspath),
        precision=precision,
        with_settings_label=with_settings_label,
    )

    """ write .ports in CSV"""
    if len(component.ports) > 0:
        with open(ports_path, "w") as fw:
            for _, port in component.ports.items():
                layer, texttype = pd._parse_layer(port.layer)

                fw.write(
                    "{}, {:.3f}, {:.3f}, {}, {:.3f}, {}, {}\n".format(
                        port.name,
                        port.x,
                        port.y,
                        int(port.orientation),
                        port.width,
                        layer,
                        texttype,
                    )
                )

    """ write JSON """
    with open(json_path, "w+") as fw:
        fw.write(json.dumps(component.get_json(), indent=2))
    return gdspath


def write_json(json_path, **settings):
    """ write properties dict into a json_path file"""

    with open(json_path, "w+") as fw:
        fw.write(json.dumps(settings))


def write_gds(
    component: Component,
    gdspath: Optional[PosixPath] = None,
    unit: float = 1e-6,
    precision: float = 1e-9,
    remove_previous_markers: bool = False,
    auto_rename: bool = False,
    with_settings_label: bool = conf.tech.with_settings_label,
) -> str:
    """ write component to GDS and returs gdspath

    Args:
        component (required)
        gdspath: by default saves it into CONFIG['gds_directory']
        auto_rename: False by default (otherwise it calls it top_cell)
        unit
        precission

    Returns:
        gdspath
    """

    gdspath = gdspath or CONFIG["gds_directory"] / (component.name + ".gds")
    gdspath = pathlib.Path(gdspath)
    gdsdir = gdspath.parent
    gdspath = str(gdspath)
    gdsdir.mkdir(parents=True, exist_ok=True)

    if remove_previous_markers:
        # If the component HAS ports AND markers and we want to
        # avoid duplicate port markers, then we clear the previous ones
        port_layer = (LAYER.PORT,)
        label_layer = (LAYER.TEXT,)
        component.remove_layers([port_layer])
        component.remove_layers([label_layer])

    # write component settings into text layer
    if with_settings_label:
        settings = component.get_settings()

        for i, k in enumerate(sorted(list(settings.keys()))):
            v = settings.get(k)
            text = f"{k} = {clean_value(v)}"
            # print(text)
            component.add_label(
                text=text,
                position=component.center - [0, i * 1],
                layer=CONFIG["layers"]["TEXT"],
            )

    component.write_gds(
        gdspath, precision=precision, auto_rename=auto_rename,
    )
    component.path = gdspath
    return gdspath


def clean_value(value):
    if isinstance(value, Component):
        value = value.name
    elif callable(value):
        value = value.__name__
    elif isinstance(value, dict):
        value = {k: clean_value(v) for k, v in value.items()}
    return value


def show(
    component: Component, gdspath: PosixPath = CONFIG["gdspath"], **kwargs
) -> None:
    """ write component GDS and shows it in klayout

    Args:
        component
        gdspath: where to save the gds
    """
    if isinstance(component, pathlib.Path):
        component = str(component)
        return klive.show(component)
    elif isinstance(component, str):
        return klive.show(component)
    elif hasattr(component, "path"):
        return klive.show(component.path)
    elif component is None:
        raise ValueError(
            "Component is None, make sure that your function returns the component"
        )

    elif isinstance(component, Component):
        write_gds(
            component, gdspath, **kwargs,
        )
        klive.show(gdspath)
    else:
        raise ValueError(
            f"Component is {type(component)}, make sure pass a Component or a path"
        )


if __name__ == "__main__":
    import pp

    # c = pp.c.waveguide(length=1.0016)  # rounds to 1.002 with 1nm precision
    # c = pp.c.waveguide(length=1.006)  # rounds to 1.005 with 5nm precision

    c = pp.c.waveguide(length=1.009)  # rounds to 1.010 with 5nm precision
    cc = pp.routing.add_fiber_array(c)
    pp.write_component(cc, precision=5e-9)
    pp.show(cc)

    print(c.settings)

    # gdspath = pp.write_component(c, precision=5e-9)
    # pp.show(gdspath)

    # cc = pp.routing.add_fiber_array(c)
    # gdspath = write_component(cc)

    # gdspath = write_component_type("ring_double_bus", overwrite=True, flatten=False)
    # gdspath = write_component_type("waveguide", length=5, overwrite=True)
    # gdspath = write_component_type("mmi1x2", width_mmi=5, overwrite=True)
    # gdspath = write_component_type("mzi2x2", overwrite=True)
    # gdspath = write_component_type("bend_circular", overwrite=True)
    # print(gdspath)
    # print(type(gdspath))
    # klive.show(gdspath)



""" there are multiple options to define layers:

- using an int
- using `pp.LAYER` object, with a layer per property
- using `pp.layer(name, layermap)` function

"""

from dataclasses import dataclass
import gdspy as gp
from phidl import LayerSet
from phidl.device_layout import DeviceReference
from phidl.device_layout import Polygon


@dataclass
class Layer:
    WG = (1, 0)
    WGCLAD = (111, 0)
    SLAB150 = (2, 0)
    SLAB90 = (3, 0)
    DEEPTRENCH = (7, 0)
    WGN = (34, 0)
    N = (20, 0)
    Np = (22, 0)
    Npp = (24, 0)
    P = (21, 0)
    Pp = (23, 0)
    Ppp = (25, 0)
    HEATER = (47, 0)
    M1 = (41, 0)
    M2 = (45, 0)
    M3 = (49, 0)
    VIA1 = (40, 0)
    VIA2 = (44, 0)
    VIA3 = (43, 0)
    NO_TILE_SI = (63, 0)
    PADDING = (67, 0)
    DEVREC = (68, 0)
    FLOORPLAN = (600, 0)
    TEXT = (66, 0)
    PORT = (1, 10)
    PORTE = (69, 0)
    PORTH = (70, 0)
    LABEL = (201, 0)
    INFO_GEO_HASH = (202, 0)
    polarization_te = (203, 0)
    polarization_tm = (204, 0)


LAYER = Layer()

# This is only for plotgds to look good
ls = LayerSet()  # Create a blank LayerSet
ls.add_layer("wgcore", 1, 0, "wgcore", "gray")
ls.add_layer("wgclad", 111, 0, "wgclad", "gray", alpha=0)
ls.add_layer("slab150", 2, 0, "slab150", "lightblue", alpha=0.2)
ls.add_layer("slab150clad", 112, 0, "slab150clad", "lightblue", alpha=0.2)
ls.add_layer("slab150trench", 122, 0, "slab150trench", "lightblue", alpha=0.2)
ls.add_layer("slab90", 3, 0, "slab90", "gray", alpha=0.5)
ls.add_layer("slab90clad", 113, 0, "slab90clad", "gray", alpha=0.5)
ls.add_layer("slab90trench", 123, 0, "slab90trench", "gray", alpha=0.5)
ls.add_layer("wgn", 34, 0, "wgn", "orange")
ls.add_layer("wgnclad", 36, 0, "wgn_clad", "gray", alpha=0)

ls.add_layer("n", 20, 0, "n", "red", alpha=0.2)
ls.add_layer("np", 22, 0, "np", "red", alpha=0.4)
ls.add_layer("npp", 24, 0, "npp", "red", alpha=0.6)

ls.add_layer("p", 21, 0, "p", "blue", alpha=0.2)
ls.add_layer("pp", 23, 0, "pp", "blue", alpha=0.4)
ls.add_layer("ppp", 25, 0, "ppp", "blue", alpha=0.6)

ls.add_layer("nbn", 31, 0, "nbn", "green", alpha=0.2)

ls.add_layer("m1", 41, 0, "m1", "green", alpha=0.2)
ls.add_layer("m2", 45, 0, "m2", "green", alpha=0.4)
ls.add_layer("m3", 49, 0, "m3", "green", alpha=0.8)
ls.add_layer("mh", 47, 0, "mh", "orange", alpha=0.2)

ls.add_layer("via1", 40, 0, "via1", "red", alpha=0.2)
ls.add_layer("via2", 40, 0, "via2", "red", alpha=0.2)
ls.add_layer("via3", 40, 0, "via3", "red", alpha=0.2)

ls.add_layer("txt", 66, 0, "txt", "grey", alpha=0.5)
ls.add_layer("label", 201, 0, "label", "grey", alpha=0)
ls.add_layer("geo_hash", 202, 0, "label", "grey", alpha=0)
ls.add_layer("te", 203, 0, "label", "grey", alpha=0)
ls.add_layer("tm", 204, 0, "label", "grey", alpha=0)
ls.add_layer("floorplan", 64, 0, "floorplan", "grey", alpha=0.2)
ls.add_layer("dicing", 65, 0, "dicing", "grey", alpha=0.2)
ls.add_layer("padding", 68, 0, "padding", "grey", alpha=0.2)
ls.add_layer("floorplan_old", 99, 0, "floorplan_old", "grey", alpha=0)
ls.add_layer("devrec", 68, 0, "devrec", "grey", alpha=0)
ls.add_layer("drcexclude", 67, 0, "drcexclude", "grey", alpha=0)
ls.add_layer("pin", 1, 10, "pin", "gray", alpha=0)
ls.add_layer("no_tile_si", 63, 0, "no_tile_si", "grey", alpha=0)
ls.add_layer("no_tile_m1", 41, 30, "no_tile_m1", "gray", alpha=0)
ls.add_layer("no_tile_m2", 45, 30, "no_tile_m2", "gray", alpha=0)
ls.add_layer("no_tile_m3", 49, 30, "no_tile_m3", "gray", alpha=0)

layer2nm = {LAYER.WG: 220}
layer2material = {
    LAYER.WG: "si",
    LAYER.SLAB90: "si",
    LAYER.SLAB150: "si",
    LAYER.WGCLAD: "sio2",
    LAYER.WGN: "sin",
}

port_layer2type = {LAYER.PORT: "optical", LAYER.PORTE: "dc", LAYER.PORTH: "heater"}

port_type2layer = {v: k for k, v in port_layer2type.items()}


def get_gds_layers(device):
    """ Returns a set of layers in this cell.

    Returns:
        out : Set of the layers used in this cell.
    """
    layers = set()
    for element in device.references:
        if isinstance(element, Polygon):
            gds_layer = (element.layers[0], element.datatypes[0])
            layers.update([gds_layer])
        elif isinstance(element, DeviceReference) or isinstance(element, gp.CellArray):
            layers.update(get_gds_layers(element.ref_cell))
    for label in device.labels:
        datatype = label.datatype if hasattr(label, "datatype") else 0
        layers.update([(label.layer, datatype)])
    return layers


def preview_layerset(ls=ls, size=100):
    """ Generates a preview Device with representations of all the layers,
    used for previewing LayerSet color schemes in quickplot or saved .gds
    files
    """
    import pp
    import numpy as np

    D = pp.Component(name="layerset")
    scale = size / 100
    num_layers = len(ls._layers)
    matrix_size = int(np.ceil(np.sqrt(num_layers)))
    sorted_layers = sorted(
        ls._layers.values(), key=lambda x: (x.gds_layer, x.gds_datatype)
    )
    for n, layer in enumerate(sorted_layers):
        R = pp.c.rectangle(size=(100 * scale, 100 * scale), layer=layer)
        T = pp.c.text(
            text="%s\n%s / %s" % (layer.name, layer.gds_layer, layer.gds_datatype),
            size=20 * scale,
            position=(50 * scale, -20 * scale),
            justify="center",
            layer=layer,
        )

        xloc = n % matrix_size
        yloc = int(n // matrix_size)
        D.add_ref(R).movex(200 * xloc * scale).movey(-200 * yloc * scale)
        D.add_ref(T).movex(200 * xloc * scale).movey(-200 * yloc * scale)
    return D


# For port labelling purpose
LAYERS_OPTICAL = [LAYER.WG]
LAYERS_ELECTRICAL = [LAYER.M1, LAYER.M2, LAYER.M3]
LAYERS_HEATER = [LAYER.HEATER]

if __name__ == "__main__":
    import pp

    c = preview_layerset(ls)
    pp.show(c)
    # print(LAYERS_OPTICAL)
    # print(layer("wgcore"))
    # print(layer("wgclad"))
    # print(layer("padding"))
    # print(layer("TEXT"))
    # print(type(layer("wgcore")))



from typing import Callable
import pp
from pp.components.coupler90 import coupler90
from pp.components.coupler_straight import coupler_straight
from pp.drc import assert_on_2nm_grid
from pp.component import Component


@pp.autoname
def coupler_ring(
    coupler90: Callable = coupler90,
    coupler: Callable = coupler_straight,
    length_x: float = 4.0,
    gap: float = 0.2,
    wg_width: float = 0.5,
    bend_radius: float = 5.0,
) -> Component:
    """ coupler for half a ring

    .. code::

           N0            N1
           |             |
            \           /
             \         /
           ---=========---
        W0    length_x    E0

    .. plot::
      :include-source:

      import pp

      c = pp.c.coupler_ring(length_x=20, bend_radius=5.0, gap=0.3, wg_width=0.45)
      pp.plotgds(c)

    """
    c = pp.Component()
    assert_on_2nm_grid(gap)

    # define subcells
    coupler90 = pp.call_if_func(
        coupler90, gap=gap, width=wg_width, bend_radius=bend_radius
    )
    coupler_straight = pp.call_if_func(
        coupler, gap=gap, length=length_x, width=wg_width
    )

    # add references to subcells
    cbl = c << coupler90
    cbr = c << coupler90
    cs = c << coupler_straight

    # connect references
    cs.connect(port="E0", destination=cbr.ports["W0"])
    cbl.reflect(p1=(0, coupler90.y), p2=(1, coupler90.y))
    cbl.connect(port="W0", destination=cs.ports["W0"])

    c.add_port("W0", port=cbl.ports["E0"])
    c.add_port("N0", port=cbl.ports["N0"])
    c.add_port("E0", port=cbr.ports["E0"])
    c.add_port("N1", port=cbr.ports["N0"])
    return c


if __name__ == "__main__":
    # c = coupler_ring(bend_radius=5.0, gap=0.3, wg_width=0.45)
    c = coupler_ring(length_x=20, bend_radius=5.0, gap=0.3, wg_width=0.45)
    # print(c.get_settings())
    print(c.name)
    pp.show(c)



import gdspy as gds
import numpy as np
import pp


def taper(start_width, end_width, length, start_coord, layer=pp.LAYER.WG, datatype=0):
    s = start_coord
    top_left = (s[0], s[1] + start_width / 2.0)
    bot_left = (s[0], s[1] - start_width / 2.0)
    top_right = (s[0] + length, s[1] + end_width / 2.0)
    bot_right = (s[0] + length, s[1] - end_width / 2.0)
    e = (s[0] + length, s[1])
    p = gds.Polygon(
        [top_left, bot_left, bot_right, top_right], layer=layer, datatype=datatype
    )
    return p, s, e


def straight(width, length, start_coord, layer=pp.LAYER.WG, datatype=0):
    t, s, e = taper(width, width, length, start_coord, layer=layer, datatype=datatype)
    return t, s, e


@pp.autoname
def spiral_circular(
    length=1e3,
    wg_width=0.5,
    spacing=3,
    min_bend_radius=5,
    points=1000,
    wg_layer=pp.LAYER.WG,
):
    """ Returns a circular spiral

    Args:
        length(um):

    .. plot::
        :include-source:

        import pp

        c = pp.c.spiral_circular(length=1e3)
        pp.plotgds(c)

    """
    wg_datatype = wg_layer[1]
    wg_layer = wg_layer[0]

    def pol_to_rect(radii, angles_deg):
        angles_rad = np.radians(angles_deg)
        z = radii * np.exp(1.0j * angles_rad)
        return z.real, z.imag

    ps = []

    # Estimate number of revolutions
    length_total = 0.0
    i = 0
    while length_total <= length:
        length_total += 3.0 * np.pi * (min_bend_radius * 2.0 + (i + 0.5) * spacing)
        i += 1
    revolutions = i + 1

    # Long spiral
    inner_revs = min_bend_radius * 4.0 / (spacing * 4.0)
    theta_1 = np.linspace(360.0 * inner_revs, 360.0 * (revolutions - 1) + 270, points)
    theta_2 = np.linspace(360.0 * inner_revs, 360.0 * revolutions, points)
    a = np.sqrt(spacing / 180.0)
    radii_1 = a ** 2 * theta_1
    radii_2 = -(a ** 2) * theta_2
    x_1, y_1 = pol_to_rect(radii_1, theta_1)
    x_1 = np.append(x_1, x_1[-1] + 0.03)
    y_1 = np.append(y_1, y_1[-1])

    p = gds.PolyPath(np.c_[x_1, y_1], wg_width, layer=wg_layer, datatype=wg_datatype)
    ps.append(p)
    x_2, y_2 = pol_to_rect(radii_2, theta_2)
    x_2 = np.append(x_2, x_2[-1])
    y_2 = np.append(y_2, y_2[-1] - 0.03)
    p = gds.PolyPath(np.c_[x_2, y_2], wg_width, layer=wg_layer, datatype=wg_datatype)
    ps.append(p)

    start_1 = (x_1[-1], y_1[-1])
    start_2 = (x_2[-1], y_2[-1])
    end_1 = (x_1[0], y_1[0])
    end_2 = (x_2[0], y_2[0])

    length_1 = np.sum(np.hypot(np.diff(x_1), np.diff(y_1)))
    length_2 = np.sum(np.hypot(np.diff(x_2), np.diff(y_2)))

    # Inner bend
    theta_inner = np.linspace(360.0 * inner_revs, 360.0 * inner_revs + 180.0, 50)
    radii_inner_1 = min_bend_radius
    radii_inner_2 = -min_bend_radius
    x_1, y_1 = pol_to_rect(radii_inner_1, theta_inner)
    x_1 -= end_1[0] / 2.0
    y_1 -= end_1[1] / 2.0
    x_1 = np.append(x_1, x_1[-1])
    y_1 = np.append(y_1, y_1[-1] - 0.03)
    p = gds.PolyPath(np.c_[x_1, y_1], wg_width, layer=wg_layer, datatype=wg_datatype)
    ps.append(p)
    x_2, y_2 = pol_to_rect(radii_inner_2, theta_inner)
    x_2 -= end_2[0] / 2.0
    y_2 -= end_2[1] / 2.0
    x_2 = np.append(x_2, x_2[-1])
    y_2 = np.append(y_2, y_2[-1] + 0.03)
    p = gds.PolyPath(np.c_[x_2, y_2], wg_width, layer=wg_layer, datatype=wg_datatype)
    ps.append(p)
    length_3 = np.sum(np.hypot(np.diff(x_2), np.diff(y_2)))

    # Output straight
    p, _, e = straight(
        wg_width, radii_1[-1], start_1, layer=wg_layer, datatype=wg_datatype
    )
    ps.append(p)

    # Outer bend
    theta_input = np.linspace(0.0, -90.0, 50)
    r_input = min_bend_radius * 2.0
    x_input, y_input = pol_to_rect(r_input, theta_input)
    x_input += start_2[0] - r_input
    y_input += start_1[1] + r_input
    s = (x_input[-1], y_input[-1])
    end_input = (x_input[0], y_input[0])
    x_input = np.append(x_input[0], x_input)
    y_input = np.append(y_input[0] + 0.03, y_input)
    x_input = np.append(x_input, x_input[-1] - 0.03)
    y_input = np.append(y_input, y_input[-1])
    p = gds.PolyPath(
        np.c_[x_input, y_input], wg_width, layer=wg_layer, datatype=wg_datatype
    )
    ps.append(p)
    length = start_2[1] - end_input[1]
    p, _, _ = straight(
        length,
        wg_width,
        (end_input[0] - wg_width / 2.0, end_input[1] + length / 2.0),
        layer=wg_layer,
        datatype=wg_datatype,
    )
    ps.append(p)

    length = length_1 + length_2 + length_3

    # Electrode ring
    # inner_radius = min_bend_radius * 2.
    # outer_radius = a**2 * theta_2[-1]
    # mid_radius = 0.5*(inner_radius + outer_radius)
    # thickness = outer_radius - mid_radius
    # r = gds.Round((0.,0.), outer_radius, inner_radius, layer=1, max_points=1000)

    ps = gds.fast_boolean(ps, None, "or")

    """ component """
    c = pp.Component()
    c.length = length
    c.add_polygon(ps, layer=wg_layer)

    c.add_port(
        name="W0",
        midpoint=(s[0], s[1]),
        orientation=180,
        layer=wg_layer,
        width=wg_width,
    )
    c.add_port(
        name="E0",
        midpoint=(e[0], e[1]),
        orientation=180,
        layer=wg_layer,
        width=wg_width,
    )
    return c


if __name__ == "__main__":
    c = spiral_circular(length=1e3, pins=True)
    print(c.ports)
    print(c.ports.keys())
    print(c.get_ports_array())
    pp.show(c)
    pp.write_gds(c)



from typing import Tuple, Optional
import numpy as np

from phidl.geometry import offset

import pp
from pp import components as pc
from pp.component import Component


@pp.autoname
def test_resistance(
    pad_size: Tuple[float] = (50.0, 50.0),
    num_squares: int = 1000,
    width: float = 1.0,
    res_layer: Optional[Tuple[int, int]] = 0,
    pad_layer: Optional[Tuple[int, int]] = None,
    gnd_layer: Optional[Tuple[int, int]] = None,
) -> Component:

    """ meander to test resistance
    from phidl.geometry

    Args:
        pad_size: Size of the two matched impedance pads (microns)
        num_squares: Number of squares comprising the resonator wire
        width: The width of the squares (microns)
        res_layer:
        pad_layer:
        gnd_layer:

    .. plot::
      :include-source:

      import pp

      c = pp.c.test_resistance()
      pp.plotgds(c)

    """

    x = pad_size[0]
    z = pad_size[1]

    # Checking validity of input
    if x <= 0 or z <= 0:
        raise ValueError("Pad must have positive, real dimensions")
    elif width > z:
        raise ValueError("Width of cell cannot be greater than height of pad")
    elif num_squares <= 0:
        raise ValueError("Number of squares must be a positive real number")
    elif width <= 0:
        raise ValueError("Width of cell must be a positive real number")

    # Performing preliminary calculations
    num_rows = int(np.floor(z / (2 * width)))
    if num_rows % 2 == 0:
        num_rows -= 1
    num_columns = num_rows - 1
    squares_in_row = (num_squares - num_columns - 2) / num_rows

    # Compensating for weird edge cases
    if squares_in_row < 1:
        num_rows = round(num_rows / 2) - 2
        squares_in_row = 1
    if width * 2 > z:
        num_rows = 1
        squares_in_row = num_squares - 2

    length_row = squares_in_row * width

    # Creating row/column corner combination structure
    T = pp.Component()
    Row = pc.rectangle(size=(length_row, width), layer=res_layer)
    Col = pc.rectangle(size=(width, width), layer=res_layer)

    T.add_ref(Row)
    col = T.add_ref(Col)
    col.move([length_row - width, -width])

    # Creating entire waveguide net
    N = pp.Component("net")
    n = 1
    for i in range(num_rows):
        if i != num_rows - 1:
            d = N.add_ref(T)
        else:
            d = N.add_ref(Row)
        if n % 2 == 0:
            d.reflect(p1=(d.x, d.ymax), p2=(d.x, d.ymin))
        d.movey(-(n - 1) * T.ysize)
        n += 1
    d = N.add_ref(Col).movex(-width)
    d = N.add_ref(Col).move([length_row, -(n - 2) * T.ysize])

    # Creating pads
    P = pp.Component()
    Pad1 = pc.rectangle(size=(x, z), layer=pad_layer)
    Pad2 = pc.rectangle(size=(x + 5, z), layer=pad_layer)
    Gnd1 = offset(Pad1, distance=-5, layer=gnd_layer)
    Gnd2 = offset(Pad2, distance=-5, layer=gnd_layer)
    pad1 = P.add_ref(Pad1).movex(-x - width)
    pad2 = P.add_ref(Pad1).movex(length_row + width)
    P.add_ref(Gnd1).center = pad1.center
    gnd2 = P.add_ref(Gnd2)
    P.add_ref(N).y = pad1.y
    gnd2.center = pad2.center
    gnd2.movex(2.5)

    return P


if __name__ == "__main__":
    c = test_resistance()
    pp.show(c)



""" test critical dimension for width and space
"""


import pp
from pp.layers import LAYER


def square_middle(side=0.5, layer=LAYER.WG):
    component = pp.Component()
    a = side / 2
    component.add_polygon([(-a, -a), (a, -a), (a, a), (-a, a)], layer=layer)
    return component


def rectangle(width, height, layer=LAYER.WG):
    component = pp.Component()
    a = width / 2
    b = height / 2
    component.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)
    return component


def triangle_middle_up(side=0.5, layer=LAYER.WG):
    component = pp.Component()
    a = side / 2
    component.add_polygon([(-a, -a), (a, -a), (0, a)], layer=layer)
    return component


def triangle_middle_down(side=0.5, layer=LAYER.WG):
    component = pp.Component()
    a = side / 2
    component.add_polygon([(-a, a), (a, a), (0, -a)], layer=layer)
    return component


CENTER_SHAPES_MAP = {
    "S": square_middle,
    "U": triangle_middle_up,
    "D": triangle_middle_down,
}


if __name__ == "__main__":
    c = square_middle()
    pp.show(c)



import pytest
from pp.components import (
    component_type2factory,
    _components,
    _containers,
    waveguide,
)


@pytest.mark.parametrize("component_type", _components)
def test_components_ports(component_type, num_regression):
    c = component_type2factory[component_type]()
    if c.ports:
        num_regression.check(c.get_ports_array())


@pytest.mark.parametrize("component_type", _components)
def test_properties_components(component_type, data_regression):
    c = component_type2factory[component_type]()
    data_regression.check(c.get_settings())


@pytest.mark.parametrize("component_type", _containers)
def test_properties_containers(component_type, data_regression):
    c = component_type2factory[component_type](component=waveguide())
    data_regression.check(c.get_settings())



import pp
from pp.layers import LAYER
from pp.components.bend_circular import bend_circular
from pp.name import autoname
from pp.component import Component


@pp.port.deco_rename_ports
@autoname
def bend_circular_heater(radius: float = 10.0, wg_width: float = 0.5) -> Component:

    theta = -90
    start_angle = 0
    angle_resolution = 2.5
    heater_to_wg_distance = 1.2
    heater_width = 0.5

    return _bend_circular_heater(
        radius=radius,
        wg_width=wg_width,
        theta=theta,
        start_angle=start_angle,
        angle_resolution=angle_resolution,
        heater_to_wg_distance=heater_to_wg_distance,
        heater_width=heater_width,
    )


def _bend_circular_heater(
    radius: float = 10,
    wg_width: float = 0.5,
    theta: int = -90,
    start_angle: int = 0,
    angle_resolution: float = 2.5,
    heater_to_wg_distance: float = 1.2,
    heater_width: float = 0.5,
) -> Component:
    """ Creates an arc of arclength ``theta`` starting at angle ``start_angle``

    Args:
        radius
        width: of the waveguide
        theta: arc length
        start_angle:
        angle_resolution
    """
    component = Component()

    wg_bend = bend_circular(
        radius=radius,
        width=wg_width,
        theta=theta,
        start_angle=start_angle,
        angle_resolution=angle_resolution,
        layer=LAYER.WG,
    ).ref((0, 0))

    a = heater_to_wg_distance + wg_width / 2 + heater_width / 2

    heater_outer = bend_circular(
        radius=radius + a,
        width=heater_width,
        theta=theta,
        start_angle=start_angle,
        angle_resolution=angle_resolution,
        layer=LAYER.HEATER,
    ).ref((0, -a))

    heater_inner = bend_circular(
        radius=radius - a,
        width=heater_width,
        theta=theta,
        start_angle=start_angle,
        angle_resolution=angle_resolution,
        layer=LAYER.HEATER,
    ).ref((0, a))

    component.add(wg_bend)
    component.add(heater_outer)
    component.add(heater_inner)

    component.absorb(wg_bend)
    component.absorb(heater_outer)
    component.absorb(heater_inner)

    i = 0

    for device in [wg_bend, heater_outer, heater_inner]:
        for port in device.ports.values():
            component.ports["{}".format(i)] = port
            i += 1

    component.info["length"] = wg_bend.info["length"]
    component.radius = radius
    component.width = wg_width
    return component


if __name__ == "__main__":
    # from pprint import pprint
    c = bend_circular_heater()
    # pprint(lys._layers)

    # c = bend_circular()
    pp.show(c)



from typing import Callable, Dict, List, Tuple
import numpy as np
import pp
from pp.name import autoname
from pp.layers import LAYER
from pp.port import Port
from pp.port import deco_rename_ports
from pp.components.waveguide import waveguide
from pp.components.hline import hline
from pp.components.electrical.tlm import tlm
from pp.components.extension import line
from pp.component import Component


@autoname
def heater(
    length: float = 10.0,
    width: float = 0.5,
    layers_heater: List[Tuple[int, int]] = [LAYER.HEATER],
) -> Component:
    """ straight heater
    """
    c = pp.Component()
    for layer in layers_heater:
        _ref = c.add_ref(hline(length=length, width=width, layer=layer))
        c.ports = _ref.ports  # Use ports from latest layer as heater ports
        c.absorb(_ref)
    return c


def add_trenches(
    c: Component,
    sstw: float = 2.0,
    trench_width: float = 0.5,
    trench_keep_out: float = 2.0,
    trenches: List[Dict[str, int]] = [
        {"nb_segments": 2, "lane": 1, "x_start_offset": 0},
        {"nb_segments": 2, "lane": -1, "x_start_offset": 0},
    ],
    layer_trench: Tuple[int, int] = LAYER.DEEPTRENCH,
) -> Component:
    """
    Add trenches to a waveguide-heater-like component
    """

    heater_width = c.settings["heater_width"]
    heater_spacing = c.settings["heater_spacing"]
    width = c.settings["width"]
    length = c.settings["length"]

    a = heater_spacing + (width + heater_width) / 2

    # Add trenches
    if trench_width and trench_width > 0:
        tko = trench_keep_out

        for trench in trenches:
            lane = trench["lane"]
            td = tko + a + (trench_width + heater_width) / 2
            y = np.sign(lane) * (td + (abs(lane) - 1) * (trench_width + tko))
            x_start_offset = trench["x_start_offset"]

            if "segments" not in trench:
                nb_segments = trench["nb_segments"]
                trench_length = (length - (nb_segments - 1) * sstw) / nb_segments
                segments = [trench_length] * nb_segments
            else:
                segments = trench["segments"]
            x = x_start_offset
            for i, trench_length in enumerate(segments):
                trench = hline(
                    length=trench_length, width=trench_width, layer=layer_trench
                )
                _trench = trench.ref(
                    port_id="W0", position=c.ports["W0"].position + (x, y)
                )
                c.add(_trench)
                c.absorb(_trench)
                x += trench_length + sstw

    return c


@autoname
def waveguide_heater(
    length: float = 10.0,
    width: float = 0.5,
    heater_width: float = 0.5,
    heater_spacing: float = 1.2,
    metal_connection: bool = True,
    sstw: float = 2.0,
    trench_width: float = 0.5,
    trench_keep_out: float = 2.0,
    trenches: List[Dict[str, int]] = [
        {"nb_segments": 2, "lane": 1, "x_start_offset": 0},
        {"nb_segments": 2, "lane": -1, "x_start_offset": 0},
    ],
    layers_heater: List[Tuple[int, int]] = [LAYER.HEATER],
    waveguide_factory: Callable = waveguide,
    layer_trench: Tuple[int, int] = LAYER.DEEPTRENCH,
) -> Component:
    """ waveguide with heater

    .. code::

        TTTTTTTTTTTTT    TTTTTTTTTTTTT <-- trench

        HHHHHHHHHHHHHHHHHHHHHHHHHHHHHH <-- heater

        ------------------------------ <-- waveguide

        HHHHHHHHHHHHHHHHHHHHHHHHHHHHHH <-- heater

        TTTTTTTTTTTTT    TTTTTTTTTTTTT <-- trench

    .. plot::
      :include-source:

      import pp

      c = pp.c.waveguide_heater()
      pp.plotgds(c)

    """
    c = Component()

    _heater = heater(length=length, width=heater_width, layers_heater=layers_heater)

    y_heater = heater_spacing + (width + heater_width) / 2
    heater_top = c << _heater
    heater_bot = c << _heater

    heater_top.movey(+y_heater)
    heater_bot.movey(-y_heater)

    wg = c << waveguide_factory(length=length, width=width)

    for i in [heater_top, heater_bot, wg]:
        c.absorb(i)

    # Add wg ports
    for p in wg.ports.values():
        c.add_port(name=p.name, port=p)

    # Add heater ports
    for p in heater_top.ports.values():
        c.add_port(name="HT" + p.name, port=p)

    for p in heater_bot.ports.values():
        c.add_port(name="HB" + p.name, port=p)

    c.settings["width"] = width
    c.settings["heater_width"] = heater_width
    c.settings["heater_spacing"] = heater_spacing
    c.settings["length"] = length
    add_trenches(
        c, sstw, trench_width, trench_keep_out, trenches, layer_trench=layer_trench
    )

    return c


@autoname
def wg_heater_connector(
    heater_ports: List[Port],
    metal_width: float = 10.0,
    tlm_layers: List[Tuple[int, int]] = [
        LAYER.VIA1,
        LAYER.M1,
        LAYER.VIA2,
        LAYER.M2,
        LAYER.VIA3,
        LAYER.M3,
    ],
) -> Component:
    """
    Connects together a pair of wg heaters and connect to a M3 port
    """

    cmp = Component()
    assert len(heater_ports) == 2
    assert (
        heater_ports[0].orientation == heater_ports[1].orientation
    ), "both ports should be facing in the same direction"
    angle = heater_ports[0].orientation
    angle = angle % 360
    assert angle in [0, 180], "angle should be 0 or 180, got {}".format(angle)

    dx = 0.0
    dy = 0.0

    angle_to_dps = {0: [(-dx, -dy), (-dx, dy)], 180: [(dx, -dy), (dx, dy)]}
    ports = heater_ports
    hw = heater_ports[0].width

    if angle in [0, 180]:
        ports.sort(key=lambda p: p.y)
    else:
        ports.sort(key=lambda p: p.x)

    _heater_to_metal = tlm(width=0.5, height=0.5, layers=tlm_layers, vias=[])

    tlm_positions = []
    for port, dp in zip(ports, angle_to_dps[angle]):
        # Extend heater
        p = port.midpoint

        # Add via/metal transitions
        tlm_pos = p + dp
        hm = _heater_to_metal.ref(position=tlm_pos)
        tlm_positions += [tlm_pos]
        cmp.add(hm)

    ss = 1 if angle == 0 else -1

    # Connect both sides with top metal
    edge_metal_piece_width = 7.0
    x = ss * edge_metal_piece_width / 2
    top_metal_layer = tlm_layers[-1]
    cmp.add_polygon(
        line(
            tlm_positions[0] + (x, -hw / 2),
            tlm_positions[1] + (x, hw / 2),
            edge_metal_piece_width,
        ),
        layer=top_metal_layer,
    )

    # Add metal port
    cmp.add_port(
        name="0",
        midpoint=0.5 * sum(tlm_positions) + (ss * edge_metal_piece_width / 2, 0),
        orientation=angle,
        width=metal_width,
        layer=top_metal_layer,
        port_type="dc",
    )

    return cmp


@deco_rename_ports
@autoname
def wg_heater_connected(
    waveguide_heater: Callable = waveguide_heater,
    wg_heater_connector: Callable = wg_heater_connector,
    tlm_layers: List[Tuple[int, int]] = [
        LAYER.VIA1,
        LAYER.M1,
        LAYER.VIA2,
        LAYER.M2,
        LAYER.VIA3,
        LAYER.M3,
    ],
    **kwargs
) -> Component:
    """
    .. plot::
      :include-source:

      import pp

      c = pp.c.wg_heater_connected()
      pp.plotgds(c)

    """
    wg_heater = waveguide_heater(**kwargs)
    # print(wg_heater.ports.keys())
    conn1 = wg_heater_connector(
        heater_ports=[wg_heater.ports["HBE0"], wg_heater.ports["HTE0"]],
        tlm_layers=tlm_layers,
    )

    conn2 = wg_heater_connector(
        heater_ports=[wg_heater.ports["HBW0"], wg_heater.ports["HTW0"]],
        tlm_layers=tlm_layers,
    )

    cmp = Component()
    for c in [wg_heater, conn1, conn2]:
        _c = cmp.add_ref(c)
        cmp.absorb(_c)

    for port_name, p in wg_heater.ports.items():
        cmp.add_port(name=port_name, port=p)

    cmp.add_port(name=1, port=conn1.ports["0"])
    cmp.add_port(name=2, port=conn2.ports["0"])

    return cmp


def _demo_waveguide_heater():
    c = waveguide_heater(width=0.5)
    pp.write_gds(c)


if __name__ == "__main__":
    # print(c.get_optical_ports())

    c = waveguide_heater(pins=True)
    # c = wg_heater_connector(heater_ports=[c.ports["HBW0"], c.ports["W0"]])
    # c = wg_heater_connected(length=100.0, width=0.5)
    print(c.ports.keys())
    for p in c.ports.values():
        print(p.name, p.port_type)

    pp.show(c)



import pp
from pp.component import Component
from typing import Tuple


@pp.autoname
def mmi2x2(
    wg_width: float = 0.5,
    width_taper: float = 0.95,
    length_taper: float = 10.0,
    length_mmi: float = 15.45,
    width_mmi: float = 2.1,
    gap_mmi: float = 0.2,
    layer: Tuple[int, int] = pp.LAYER.WG,
) -> Component:
    """ mmi 2x2

    Args:
        wg_width: input waveguides width
        width_taper: interface between input waveguides and mmi region
        length_taper: into the mmi region
        length_mmi: in x direction
        width_mmi: in y direction
        gap_mmi: (width_taper + gap between tapered wg)/2
        layer: gds layer

    .. plot::
      :include-source:

      import pp
      c = pp.c.mmi2x2(length_mmi=15.45, width_mmi=2.1)
      pp.plotgds(c)

    """
    component = pp.Component()
    w_mmi = width_mmi
    w_taper = width_taper

    taper = pp.c.taper(length=length_taper, width1=wg_width, width2=w_taper, pins=False)

    a = gap_mmi / 2 + width_taper / 2
    mmi = pp.c.rectangle(
        size=(length_mmi, w_mmi),
        layer=layer,
        ports_parameters={
            "E": [(w_mmi / 2 - a, w_taper), (w_mmi / 2 + a, w_taper)],
            "W": [(w_mmi / 2 - a, w_taper), (w_mmi / 2 + a, w_taper)],
        },
    )

    mmi_section = component.add_ref(mmi)

    # For each port on the MMI rectangle
    for port_name, port in mmi.ports.items():

        # Create a taper
        taper_ref = component.add_ref(taper)

        # Connect the taper to the mmi section
        taper_ref.connect(port="2", destination=mmi_section.ports[port_name])

        # Add the taper port
        component.add_port(name=port_name, port=taper_ref.ports["1"])
        component.absorb(taper_ref)

    component.simulation_settings = dict(port_width=1.5e-6)
    component.absorb(mmi_section)
    return component


@pp.autoname
def mmi2x2_biased(
    wg_width=0.5,
    width_taper=0.95,
    length_taper=10,
    length_mmi=15.45,
    width_mmi=2.1,
    gap_mmi=0.2,
    layer=pp.LAYER.WG,
):
    return mmi2x2(
        wg_width=pp.bias.width(wg_width),
        width_taper=pp.bias.width(width_taper),
        length_taper=length_taper,
        length_mmi=length_mmi,
        width_mmi=pp.bias.width(width_mmi),
        gap_mmi=pp.bias.gap(gap_mmi),
        layer=layer,
    )


def test_mmi2x2():
    c = mmi2x2()
    pp.write_gds(c)


if __name__ == "__main__":
    c = mmi2x2()
    # c = mmi2x2_biased()
    # pp.write_to_libary("mmi1x2", width_mmi=10, overwrite=True)
    # print(c.get_optical_ports())
    print(c.get_settings())
    pp.show(c)



from pp.config import CONFIG


def width(width):
    """ return wg_width with bias """
    return width + CONFIG.get("bias", 0)


def gap(gap):
    """ return wg_gap with bias """
    return gap - CONFIG.get("bias", 0)


if __name__ == "__main__":
    # print(width(0.5))
    print(gap(0.5))



"""
A GDS diff tool which can be called by git when doing

`git diff ...`

The function needs to take the arguments as described in
https://git-scm.com/docs/git/2.18.0#git-codeGITEXTERNALDIFFcode

"""
import sys

from gdsdiff import gdsdiff
import pp


def gds_diff_git(
    path, curr_file, old_file, old_hex, old_mode, new_file, new_hex, new_mode
):
    """
    We do not use most of the arguments
    """
    print(old_hex, "->", new_hex)
    diff = gdsdiff.gdsdiff(old_file, new_file)
    pp.show(diff)


if __name__ == "__main__":
    # for f in sys.argv:
    # print (f)
    gds_diff_git(*sys.argv)



import math

# Some global settings
PADDING = 10000  # 10um padding
STEP_SIZE = 1
GRID = 1
WORKING_MEMORY = {}
DEVREC_LAYER = 68
FLOORPLAN_LAYER = 64
DICING_LAYERS = [(1, 0), (41, 30), (45, 30), (49, 30)]
TEXT = 66
COUNTER = 0
AUTOPLACER_REGISTRY = {}

VERTICAL = 101
HORIZONTAL = 102

NORTH = 201
SOUTH = 202
EAST = 203
WEST = 204
MIDDLE = 0

HEIGHT = 301
WIDTH = 302
BOTH = 303
AREA = 304
NEITHER = None

GRANULARITY = PADDING * 8

NORTH_EAST = (NORTH, EAST)
NORTH_WEST = (NORTH, WEST)
SOUTH_EAST = (SOUTH, EAST)
SOUTH_WEST = (SOUTH, WEST)
CENTER = (MIDDLE, MIDDLE)
CORNERS = [NORTH_EAST, NORTH_WEST, SOUTH_EAST, SOUTH_WEST]


class CellsNotFoundError(Exception):
    pass


class OutOfSpaceError(Exception):
    pass


def mean(values):
    """ Just compute the mean """
    return sum(values) / len(values)


def area(cell):
    """ Get the area of a cell, used for sorting """
    return cell.bbox().width()


def factors(n):
    """ Get all the factors of n """
    return [i for i in range(1, n + 1) if n % i == 0]


def estimate_cols(n, width, height, target):
    """
    Estimate the number of cols required
    to acheive a certain target aspect ratio
    """
    cols = factors(n)

    # Hack to deal with prime numbers
    if cols == [1, n]:
        cols = [math.ceil(math.sqrt(n))]
    aspect = {c: c * width / (height * n / c) for c in cols}
    return sorted(cols, key=lambda c: abs(aspect[c] - target))[0]


def longest_common_prefix(names):
    """ Used to infer group names """
    if not names:
        return "Empty"
    length = max(len(n) for n in names)
    for i in range(length, 0, -1):
        prefix = names[0][:i]
        if all(n.startswith(prefix) for n in names):
            return prefix
        return prefix



"""" add electrical labels to each device port

"""

import phidl.device_layout as pd
import pp
from omegaconf.listconfig import ListConfig
from phidl.device_layout import Label
from pp.component import Component, ComponentReference
from pp.port import Port
from typing import Optional, Union


def add_label(component, text, position=(0, 0), layer=pp.LAYER.LABEL):
    gds_layer_label, gds_datatype_label = pd._parse_layer(layer)
    label = pd.Label(
        text=text,
        position=position,
        anchor="o",
        layer=gds_layer_label,
        texttype=gds_datatype_label,
    )
    component.add(label)
    return component


def add_labels(component):
    c = pp.Component()
    electrical_ports = component.get_ports_list(port_type="dc")
    c.add(component.ref())

    for i, port in enumerate(electrical_ports):
        label = get_input_label_electrical(port, i, component_name=component.name)
        c.add(label)

    return c


def get_optical_text(
    port: Port,
    gc: Union[ComponentReference, Component],
    gc_index: Optional[int] = None,
    component_name: Optional[str] = None,
) -> str:
    polarization = gc.get_property("polarization")
    wavelength_nm = gc.get_property("wavelength")

    assert polarization in [
        "te",
        "tm",
    ], f"Not valid polarization {polarization} in [te, tm]"
    assert (
        isinstance(wavelength_nm, (int, float)) and 1000 < wavelength_nm < 2000
    ), f"{wavelength_nm} is Not valid 1000 < wavelength < 2000"

    if component_name:
        name = component_name

    elif type(port.parent) == pp.Component:
        name = port.parent.name
    else:
        name = port.parent.ref_cell.name

    if isinstance(gc_index, int):
        text = (
            f"opt_{polarization}_{int(wavelength_nm)}_({name})_{gc_index}_{port.name}"
        )
    else:
        text = f"opt_{polarization}_{int(wavelength_nm)}_({name})_{port.name}"

    return text


def get_input_label(
    port: Port,
    gc: ComponentReference,
    gc_index: Optional[int] = None,
    gc_port_name: str = "W0",
    layer_label: ListConfig = pp.LAYER.LABEL,
    component_name: Optional[str] = None,
) -> Label:
    """
    Generate a label with component info for a given grating coupler.
    This is the label used by T&M to extract grating coupler coordinates
    and match it to the component.
    """
    text = get_optical_text(
        port=port, gc=gc, gc_index=gc_index, component_name=component_name
    )

    if gc_port_name is None:
        gc_port_name = list(gc.ports.values())[0].name

    layer, texttype = pd._parse_layer(layer_label)
    label = pd.Label(
        text=text,
        position=gc.ports[gc_port_name].midpoint,
        anchor="o",
        layer=layer,
        texttype=texttype,
    )
    return label


def get_input_label_electrical(
    port, index=0, component_name=None, layer_label=pp.LAYER.LABEL
):
    """
    Generate a label to test component info for a given grating coupler.
    This is the label used by T&M to extract grating coupler coordinates
    and match it to the component.
    """

    if component_name:
        name = component_name

    elif type(port.parent) == pp.Component:
        name = port.parent.name
    else:
        name = port.parent.ref_cell.name

    text = "elec_{}_({})_{}".format(index, name, port.name)

    layer, texttype = pd._parse_layer(layer_label)

    label = pd.Label(
        text=text, position=port.midpoint, anchor="o", layer=layer, texttype=texttype,
    )
    return label


def _demo_input_label():
    c = pp.c.bend_circular()
    gc = pp.c.grating_coupler_elliptical_te()
    label = get_input_label(port=c.ports["W0"], gc=gc, layer_label=pp.LAYER.LABEL)
    print(label)


if __name__ == "__main__":
    from pp.components.electrical.pad import pad

    c = pad(width=10, height=10)
    print(c.ports)
    c2 = add_labels(c)
    pp.show(c2)



from phidl.device_layout import _parse_layer
from pp.geo_utils import polygon_grow


def add_keepout(c, target_layers, keepout_layers, margin=2.0):
    """ Lookup all polygons in this cell. You can also use add_padding
    """
    all_cells = list(c.get_dependencies(True)) + [c]
    for _c in all_cells:
        for _layer in target_layers:
            polys_by_layer = _c.get_polygons(by_spec=True)
            layer = _parse_layer(_layer)
            for ko_layer in keepout_layers:
                ko_layer = _parse_layer(ko_layer)
                polys = polys_by_layer[layer] if layer in polys_by_layer else []
                if not polys:
                    continue

                ko_polys_pts = [polygon_grow(poly, margin) for poly in polys]
                _c.add_polygon(ko_polys_pts, ko_layer)


if __name__ == "__main__":
    from pp.components.crossing_waveguide import crossing_etched
    from pp.components.crossing_waveguide import crossing45
    from pp.layers import LAYER

    c = crossing45(alpha=0.5, crossing=crossing_etched)

    target_layers = [LAYER.WG, LAYER.SLAB150]
    keepout_layers = [LAYER.NO_TILE_SI]

    add_keepout(c, target_layers, keepout_layers)
    import pp

    pp.show(c)



import numpy as np
from omegaconf.listconfig import ListConfig
from pp.component import Component
from typing import List, Tuple

import pp
from pp.layers import LAYER
from pp.name import clean_name

"""
A pixel based font, guaranteed to be manhattan, without accute angles
"""


def manhattan_text(
    text: str = "abcd",
    size: float = 10,
    position: Tuple[int, int] = (0, 0),
    justify: str = "left",
    layer: ListConfig = LAYER.M1,
    layers_cladding: List[ListConfig] = [],
    cladding_offset: int = 3,
) -> Component:
    """

    .. plot::
      :include-source:

      import pp

      c = pp.c.text(text="abcd", size=10, position=(0, 0), justify="left", layer=1)
      pp.plotgds(c)

    """
    pixel_size = size
    xoffset = position[0]
    yoffset = position[1]
    t = pp.Component(
        name=clean_name(text) + "_{}_{}".format(int(position[0]), int(position[1]))
    )
    for i, line in enumerate(text.split("\n")):
        component = pp.Component(name=t.name + "{}".format(i))
        for c in line:
            try:
                if c not in CHARAC_MAP:
                    c = c.upper()
                pixels = CHARAC_MAP[c]
            except:
                print(
                    "character {} could not be written (probably not part of dictionnary)".format(
                        c
                    )
                )
                continue

            _c = component.add_ref(
                pixel_array(pixels=pixels, pixel_size=pixel_size, layer=layer)
            )
            _c.move((xoffset, yoffset))
            component.absorb(_c)
            xoffset += pixel_size * 6

        t.add_ref(component)
        yoffset -= pixel_size * 6
        xoffset = position[0]
    justify = justify.lower()
    for ref in t.references:
        if justify == "left":
            pass
        if justify == "right":
            ref.xmax = position[0]
        if justify == "center":
            ref.move(origin=ref.center, destination=position, axis="x")

    points = [
        [t.xmin - cladding_offset / 2, t.ymin - cladding_offset],
        [t.xmax + cladding_offset / 2, t.ymin - cladding_offset],
        [t.xmax + cladding_offset / 2, t.ymax + cladding_offset],
        [t.xmin - cladding_offset / 2, t.ymax + cladding_offset],
    ]
    for layer in layers_cladding:
        t.add_polygon(points, layer=layer)
    return t


@pp.autoname
def pixel_array(
    pixels: str = """
     XXX
    X   X
    XXXXX
    X   X
    X   X
    """,
    pixel_size: float = 10.0,
    layer: ListConfig = LAYER.M1,
) -> Component:
    component = pp.Component()
    lines = [line for line in pixels.split("\n") if len(line) > 0]
    lines.reverse()
    j = 0
    i = 0
    i_max = 0
    a = pixel_size
    for line in lines:
        i = 0
        for c in line:
            if c in ["X", "1"]:
                p0 = np.array([i * a, j * a])
                pixel = [p0 + p for p in [(0, 0), (a, 0), (a, a), (0, a)]]
                component.add_polygon(pixel, layer=layer)
            i += 1
        i_max = max(i_max, i)
        j += 1

    return component


FONT = """\
A
1	1	1	1	1
1	0	0	0	1
1	1	1	0	1
1	0	0	0	1
1	0	0	0	1
B
1	1	1	1	1
1	0	0	0	1
1	0	1	1	1
1	0	0	0	1
1	0	1	1	1
C
1	1	1	1	1
1	0	0	0	1
1	0	0	0	0
1	0	0	0	1
1	1	1	1	1
D
1	1	1	1	1
1	0	0	0	1
1	0	0	0	1
1	0	0	0	1
1	0	1	1	1
E
1	1	1	1	1
1	0	0	0	0
1	0	1	1	0
1	0	0	0	0
1	0	1	1	1
F
1	1	1	1	1
1	0	0	0	0
1	0	1	1	0
1	0	0	0	0
1	0	0	0	0
G
1	1	1	1	1
1	0	0	0	0
1	0	1	1	1
1	0	0	0	1
1	1	1	1	1
H
1	0	0	0	1
1	0	0	0	1
1	1	1	1	1
1	0	0	0	1
1	0	0	0	1
I
1	1	1	1	1
0	0	1	0	0
0	0	1	0	0
0	0	1	0	0
1	1	1	1	1
J
0	0	0	0	1
0	0	0	0	1
0	0	0	0	1
1	0	0	0	1
1	1	1	1	1
K
1	0	0	0	1
1	0	0	1	1
1	1	1	1	0
1	0	0	1	1
1	0	0	0	1
L
1	0	0	0	0
1	0	0	0	0
1	0	0	0	0
1	0	0	0	0
1	1	1	1	1
M
1	1	0	1	1
1	1	1	1	1
1	0	1	0	1
1	0	0	0	1
1	0	0	0	1
N
1	1	1	0	1
1	0	1	0	1
1	0	1	1	1
1	0	0	1	1
1	0	0	0	1
O
1	1	1	1	1
1	0	0	0	1
1	0	0	0	1
1	0	0	0	1
1	1	0	1	1
P
1	1	1	1	1
1	0	0	0	1
1	0	1	1	1
1	0	0	0	0
1	0	0	0	0
Q
1	1	1	1	0
1	0	0	1	0
1	0	0	1	0
1	0	0	1	0
1	1	1	1	1
R
1	1	1	1	0
1	0	0	1	0
1	0	1	1	1
1	0	0	0	1
1	0	0	0	1
S
1	1	1	1	1
1	1	0	0	0
0	1	1	1	0
0	0	0	1	1
1	1	1	1	1
T
1	1	1	1	1
0	0	1	0	0
0	0	1	0	0
0	0	1	0	0
0	0	1	0	0
U
1	0	0	0	1
1	0	0	0	1
1	0	0	0	1
1	0	0	0	1
1	1	1	1	1
V
1	0	0	0	1
1	0	0	0	1
1	0	0	0	1
0	1	0	1	0
0	0	1	0	0
W
1	0	0	0	1
1	0	0	0	1
1	0	1	0	1
1	1	1	1	1
1	1	0	1	1
X
1	1	0	1	1
1	1	1	1	1
0	1	1	1	0
1	1	1	1	1
1	1	0	1	1
Y
1	0	0	0	1
1	0	0	0	1
1	1	1	1	1
0	0	1	0	0
0	0	1	0	0
Z
1	1	1	1	1
0	0	0	1	1
0	0	1	1	0
0	1	1	0	0
1	1	1	1	1
1
0	1	1	0	0
0	0	1	0	0
0	0	1	0	0
0	0	1	0	0
0	1	1	1	0
2
1	1	1	1	1
0	0	0	1	1
1	1	1	1	1
1	1	0	0	0
1	1	1	1	1
3
1	1	1	1	1
0	0	0	1	1
1	1	1	1	1
0	0	0	1	1
1	1	1	1	1
4
1	0	0	1	1
1	0	0	1	1
1	1	1	1	1
0	0	0	1	1
0	0	0	1	1
5
1	1	1	1	1
1	1	0	0	0
1	1	1	1	1
0	0	0	1	1
1	1	1	1	1
6
1	1	1	1	1
1	0	0	0	0
1	1	1	1	1
1	1	0	1	1
1	1	1	1	1
7
1	1	1	1	1
0	0	0	1	1
0	0	0	1	1
0	0	0	1	1
0	0	0	1	1
8
1	1	1	1	1
1	1	0	1	1
1	1	1	1	1
1	1	0	1	1
1	1	1	1	1
9
1	1	1	1	1
1	1	0	1	1
1	1	1	1	1
0	0	0	1	1
0	0	0	1	1
0
0	1	1	1	1
0	1	0	0	1
0	1	0	0	1
0	1	0	0	1
0	1	1	1	1
+
0	0	0	0	0
0	0	1	0	0
0	1	1	1	0
0	0	1	0	0
0	0	0	0	0
-
0	0	0	0	0
0	0	0	0	0
0	1	1	1	0
0	0	0	0	0
0	0	0	0	0
_
0	0	0	0	0
0	0	0	0	0
0	0	0	0	0
0	0	0	0	0
0	1	1	1	0
.
0	0	0	0	0
0	0	0	0	0
0	0	0	0	0
0	0	0	0	0
0	0	1	0	0

0	0	0	0	0
0	0	0	0	0
0	0	0	0	0
0	0	0	0	0
0	0	0	0	0
"""

CHARAC_MAP = {}


def load_font() -> None:
    lines = FONT.split("\n")
    global CHARAC_MAP
    while lines:
        line = lines.pop(0)
        if not line:
            break
        charac = line[0]

        pixels = ""
        for i in range(5):
            pixels += lines.pop(0).replace("\t", "").replace(" ", "") + "\n"

        CHARAC_MAP[charac] = pixels


load_font()


if __name__ == "__main__":
    c = manhattan_text(
        text="The mask is nearly done. only 12345 drc errors remaining",
        layers_cladding=[(33, 44)],
    )
    pp.show(c)



import pp
from pp.component_from_yaml import component_from_yaml
from pp.component import Component
from pp.name import autoname


@autoname
def coupler(
    wg_width: float = 0.5, gap: float = 0.236, length: float = 10.007
) -> Component:

    pp.drc.assert_on_1nm_grid(length)
    pp.drc.assert_on_2nm_grid(gap)

    netlist = f"""
instances:
    sl:
      component: coupler_symmetric
      settings:
        gap: {gap}
        wg_width: {wg_width}
    sr:
      component: coupler_symmetric
      settings:
        gap: {gap}
        wg_width: {wg_width}
    cs:
      component: coupler_straight
      settings:
        gap: {gap}
        width: {wg_width}
        length: {length}

placements:
    cs:
        x: 0
        y: {gap/2}

connections:
    sl,W0: cs,W0
    sr,W0: cs,E0

ports:
    w1: sl,E0
    w0: sl,E1
    e0: sr,E0
    e1: sr,E1

    """
    return component_from_yaml(netlist)


if __name__ == "__main__":
    c = coupler(pins=True)
    pp.show(c)
